import asyncio
import re
from contextlib import asynccontextmanager
from typing import Any

from core import (
    Database,
    UserContext,
    build_user_context,
    create_agent,
)
from fastapi import FastAPI, HTTPException, Request
from prefect import flow, get_run_logger, task
from prefect.blocks.notifications import SlackWebhook
from prefect.cache_policies import NONE
from prefect.logging.loggers import get_logger
from prefect.states import Completed
from prefect.variables import Variable
from pydantic_ai.agent import AgentRunResult
from pydantic_ai.messages import ModelMessage
from settings import settings
from slack import SlackPayload, get_channel_name, post_slack_message
from strings import count_tokens, slice_tokens
from wrap import WatchToolCalls

BOT_MENTION = r"<@(\w+)>"


logger = get_logger(__name__)


@task(name="run agent loop")
async def run_agent(
    cleaned_message: str,
    conversation: list[ModelMessage],
    user_context: UserContext,
    decorator_settings: dict[str, Any] | None = None,
) -> AgentRunResult[str]:
    if decorator_settings is None:
        decorator_settings = {
            "cache_policy": NONE,
            "task_run_name": "execute {self.function.__name__}",
            "log_prints": True,
        }

    with WatchToolCalls(settings=decorator_settings):
        result = await create_agent(model="openai:gpt-4o").run(
            user_prompt=cleaned_message,
            message_history=conversation,
            deps=user_context,
        )
        return result


@flow(name="Handle Slack Message", retries=1)
async def handle_message(payload: SlackPayload, db: Database):
    logger = get_run_logger()
    event = payload.event
    if not event or not all([event.text, event.channel, (event.thread_ts or event.ts)]):
        logger.debug("Skipping invalid event")
        return Completed(message="Invalid event", name="SKIPPED")

    USER_MESSAGE_MAX_TOKENS = settings.user_message_max_tokens
    user_message = event.text or ""
    thread_ts = event.thread_ts or event.ts
    assert thread_ts is not None, "No thread_ts found"
    cleaned_message = re.sub(BOT_MENTION, "", user_message).strip()
    msg_len = count_tokens(cleaned_message)

    if msg_len > USER_MESSAGE_MAX_TOKENS:
        logger.warning(
            f"Message too long by {msg_len - USER_MESSAGE_MAX_TOKENS} tokens"
        )
        assert event.channel is not None, "No channel found"
        await post_slack_message(
            message=(
                "Your message was too long, here's your message at the allowed limit:"
                f"\n{slice_tokens(cleaned_message, USER_MESSAGE_MAX_TOKENS)}"
            ),
            channel_id=event.channel,
            thread_ts=thread_ts,
        )
        return Completed(message="Message too long", name="SKIPPED")

    if re.search(BOT_MENTION, user_message) and payload.authorizations:
        logger.info(
            f"Processing message in thread {thread_ts}\nUser message: {cleaned_message}"
        )
        conversation = await db.get_thread_messages(thread_ts)

        user_context = build_user_context(
            user_id=payload.authorizations[0].user_id,
            user_question=cleaned_message,
        )

        result = await run_agent(cleaned_message, conversation, user_context)  # type: ignore

        await db.add_thread_messages(thread_ts, result.new_messages())
        assert event.channel is not None, "No channel found"
        await task(post_slack_message)(
            message=result.data,
            channel_id=event.channel,
            thread_ts=thread_ts,
        )
        return Completed(message="Responded to mention")

    return Completed(message="Skipping non-mention", name="SKIPPED")


@asynccontextmanager
async def lifespan(app: FastAPI):
    async with Database.connect(settings.db_file) as db:
        app.state.db = db
        yield


app = FastAPI(lifespan=lifespan)


@app.post("/chat")
async def chat_endpoint(request: Request) -> dict[str, Any]:
    try:
        payload = SlackPayload(**await request.json())
    except Exception as e:
        logger.error(f"Error parsing Slack payload: {e}")
        slack_webhook = await SlackWebhook.load("marvin-bot-pager")  # type: ignore
        await slack_webhook.notify(  # type: ignore
            body=f"Error parsing Slack payload: {e}",
            subject="Slackbot Error",
        )
        raise HTTPException(400, "Invalid event type")

    db: Database = request.app.state.db

    if payload.type == "event_callback":
        if not payload.event:
            raise HTTPException(400, "No event found")

        if payload.event.type == "team_join":
            logger.info(f"New team member joined: {payload.event.user}")
            user_id = payload.event.user
            assert isinstance(user_id, str), "expected user_id to be a string"
            message_variable = await Variable.aget("marvin_welcome_message")
            message_text = message_variable.value["text"]  # type: ignore
            assert isinstance(message_text, str), "expected message_text to be a string"
            welcome_text = message_text.format(user_id=user_id)
            await post_slack_message(
                welcome_text,
                channel_id=user_id,
            )
        else:
            if not payload.event.channel:
                raise HTTPException(400, "No channel found")

            channel_name = await get_channel_name(payload.event.channel)
            if channel_name.startswith("D"):
                logger.warning(f"Attempted DM in channel: {channel_name}")
                slack_webhook = await SlackWebhook.load("marvin-bot-pager")  # type: ignore
                await slack_webhook.notify(  # type: ignore
                    body=f"Attempted DM: {channel_name}",
                    subject="Slackbot DM Warning",
                )
                return {"status": "skipped dm"}

            logger.info(f"Processing message in {channel_name}")
            ts = payload.event.thread_ts or payload.event.ts
            flow_opts: dict[str, Any] = dict(
                flow_run_name=f"respond in {channel_name}/{ts}",
            )

            asyncio.create_task(handle_message.with_options(**flow_opts)(payload, db))
    elif payload.type == "url_verification":
        return {"challenge": payload.challenge}
    else:
        raise HTTPException(400, "Invalid event type")

    return {"status": "ok"}



import os
from pathlib import Path

import httpx
from pydantic_ai.models.openai import OpenAIModel
from pydantic_ai.providers.openai import OpenAIProvider

import marvin


def write_file(path: str, content: str):
    """Write content to a file"""
    _path = Path(path)
    _path.write_text(content)


writer = marvin.Agent(
    model=OpenAIModel(
        "gpt-4o",
        provider=OpenAIProvider(
            api_key=os.getenv("OPENAI_API_KEY"),
            base_url="https://api.openai.com/v1",
            http_client=httpx.AsyncClient(
                # proxy="http://localhost:8080",
                # headers={"x-SOME-HEADER": "some-value"},
            ),
        ),
    ),
    name="Technical Writer",
    instructions="Write concise, engaging content for developers",
    tools=[write_file],
)

result = marvin.run("how to use pydantic? write haiku to docs.md", agents=[writer])
print(result)

"""
» uv run examples/hello_agent.py
╭─ Agent "Technical Writer" (d9cf5814) ────────────────────────────────────────────────────────────╮
│                                                                                                  │
│  Tool:    write_file                                                                             │
│  Status:  ✅                                                                                     │
│  Input    {                                                                                      │
│               'path': 'docs.md',                                                                 │
│               'content': '### Pydantic Haiku\n\nModel your data.\nValidation made                │
│           easy.\nPydantic shines bright.\n'                                                      │
│           }                                                                                      │
│                                                                                                  │
╰────────────────────────────────────────────────────────────────────────────────────  4:35:45 PM ─╯
╭─ Agent "Technical Writer" (d9cf5814) ────────────────────────────────────────────────────────────╮
│                                                                                                  │
│  Tool:    Mark Task 8b1894c6 ("how to use pydantic? write haiku to docs...") successful          │
│  Status:  ✅                                                                                     │
│  Result   Pydantic haiku added to docs.md                                                        │
│                                                                                                  │
╰────────────────────────────────────────────────────────────────────────────────────  4:35:48 PM ─╯
"""



import asyncio
import signal
import sys
from pathlib import Path
from typing import Any, ClassVar

import httpx
from prompt_toolkit import PromptSession
from prompt_toolkit.auto_suggest import AutoSuggestFromHistory
from prompt_toolkit.history import FileHistory
from pydantic import Field, SecretStr
from pydantic_ai.models.openai import OpenAIModel
from pydantic_ai.providers.deepseek import DeepSeekProvider
from pydantic_settings import BaseSettings, SettingsConfigDict

from marvin import Agent, Thread
from marvin.settings import settings as marvin_settings

running: bool = True


def handle_sigint(signum: int, frame: Any) -> None:
    global running
    running = False


signal.signal(signal.SIGINT, handle_sigint)


class Settings(BaseSettings):
    model_config: ClassVar[SettingsConfigDict] = SettingsConfigDict(
        env_file=".env", extra="ignore"
    )

    deepseek_api_key: SecretStr = Field(default=...)
    google_api_key: SecretStr = Field(default=...)
    google_cx: SecretStr = Field(default=...)
    notes_file_path: str = Field(default="notes.txt")


settings = Settings()  # type: ignore # google_cx loaded from .env file


def google_search(query: str, num: int = 3) -> str:
    """Use google to search the internet.

    Args:
        query: The query to search for.
        num: The number of results to return (preferably 3)

    Returns:
        The results of the search.
    """
    response = httpx.get(
        "https://www.googleapis.com/customsearch/v1",
        params={
            "q": query,
            "key": settings.google_api_key.get_secret_value(),
            "cx": settings.google_cx.get_secret_value(),
            "num": num,
        },
    )
    response.raise_for_status()
    return response.json()


def write_to_file(content: str, file_path: str) -> None:
    """Write content to a file.

    Args:
        content: The content to write to the file
        file_path: The path to the file to write to (will be casted to a Path)
    """
    Path(file_path).write_text(content)


async def get_user_input(session: PromptSession[str]) -> str:
    try:
        return await session.prompt_async(
            "you ➤ ",
            auto_suggest=AutoSuggestFromHistory(),
        )
    except (KeyboardInterrupt, EOFError):
        global running
        running = False
        return ""


async def main(model: str | None = None) -> int:
    try:
        history_file = marvin_settings.home_path / ".deepseek-history.txt"
        session = PromptSession[str](history=FileHistory(str(history_file)))

        agent = Agent(
            name="deepseek assistant",
            model=OpenAIModel(
                model or "deepseek-chat",
                provider=DeepSeekProvider(
                    api_key=settings.deepseek_api_key.get_secret_value(),
                ),
            ),
            tools=[google_search, write_to_file],
            prompt=f"""You are a helpful assistant that can search the internet for information.
            You can write notes in {settings.notes_file_path}.

            Find out what the user wants to do and help them with their request.
            """,
        )

        with Thread():
            while running:
                if not (user_input := await get_user_input(session)) or not running:
                    break
                if user_input.lower() in ("exit", "quit", ":q!"):
                    break
                await agent.run_async(user_input)
    except Exception as e:
        print(f"\nError: {e}")
        return 1
    finally:
        print("Goodbye!")

    return 0


if __name__ == "__main__":
    sys.exit(asyncio.run(main(sys.argv[1] if len(sys.argv) > 1 else None)))



import re
import sys

BREAKPOINT_PATTERNS = [
    r"breakpoint\(\)",
    r"pdb\.set_trace\(\)",
    r"import pdb",
    r"from pdb import",
    r"ipdb\.set_trace\(\)",
    r"import ipdb",
    r"from ipdb import",
]


def check_file(filename: str) -> list[str]:
    with open(filename, "r") as f:
        content = f.read()

    found_breakpoints: list[str] = []
    for pattern in BREAKPOINT_PATTERNS:
        matches = re.findall(pattern, content)
        for match in matches:
            found_breakpoints.append(f"{filename}: {match}")

    return found_breakpoints


if __name__ == "__main__":
    files = sys.argv[1:]
    all_breakpoints: list[str] = []

    for file in files:
        breakpoints = check_file(file)
        all_breakpoints.extend(breakpoints)

    if all_breakpoints:
        print("Found breakpoints in files:")
        for bp in all_breakpoints:
            print(f"  {bp}")
        sys.exit(1)

    sys.exit(0)



"""Warning: this example will run untrusted shell commands.

Use with caution.
"""

import platform
import subprocess

from pydantic import IPvAnyAddress

import marvin


def run_shell_command(command: list[str]) -> str:
    """e.g. ['ls', '-l'] or ['git', 'diff', '|', 'grep', 'some_code']"""
    return subprocess.check_output(command).decode()


task = marvin.Task[IPvAnyAddress](
    instructions="find the current ip address",
    result_type=IPvAnyAddress,
    tools=[run_shell_command],
    context={"os": platform.system()},
)

task.run()



import importlib
import inspect
import pkgutil
from types import ModuleType
from typing import Any

from prefect.utilities.importtools import import_object


class ModuleTreeExplorer:
    def __init__(self, root_module_path: str, max_depth: int = 2):
        """
        Initialize the module tree explorer with a root module path.

        Args:
            root_module_path: String representing the root module (e.g., 'prefect.runtime')
            max_depth: Maximum depth to explore in the module tree (default: 2)
        """
        self.root_module_path = root_module_path
        self.max_depth = max_depth
        self.tree = {}

    def _import_module(self, module_path: str) -> ModuleType | None:
        """Safely import a module.

        Args:
            module_path: String representing the module path (e.g., 'prefect.runtime')
        """
        try:
            return importlib.import_module(module_path)
        except (ImportError, TypeError) as e:
            print(f"Warning: Could not import {module_path}: {e}")
            return None

    def _is_defined_in_module(self, item: Any, current_module: str) -> bool:
        """Check if an item is defined in the current module.

        Args:
            item: The item to check
            current_module: The current module path
        """
        try:
            if inspect.ismodule(item):
                return False

            # Get the module where this item is defined
            if hasattr(item, "__module__"):
                return item.__module__ == current_module

            return False
        except Exception:
            return False

    def _get_module_public_api(
        self, module: ModuleType, module_path: str | None = None
    ) -> dict[str, list[str]]:
        """Get the public API of a module.

        Args:
            module: The module to get the public API of
            module_path: The path of the module (e.g., 'prefect.runtime')
        """
        api = {"all": [], "classes": [], "functions": [], "constants": []}

        try:
            if hasattr(module, "__all__"):
                api["all"] = list(module.__all__)
                items = {
                    name: getattr(module, name, None)
                    for name in module.__all__
                    if hasattr(module, name)
                }
            else:
                # Get non-underscore attributes
                items = {
                    name: getattr(module, name, None)
                    for name in dir(module)
                    if not name.startswith("_")
                }

            # Categorize items that we can safely inspect and are defined in our module
            module_name = module_path or module.__name__
            for name, item in items.items():
                try:
                    if item is not None and self._is_defined_in_module(
                        item, module_name
                    ):
                        if inspect.isclass(item):
                            api["classes"].append(name)
                        elif inspect.isfunction(item):
                            api["functions"].append(name)
                        elif not inspect.ismodule(item):
                            api["constants"].append(name)
                except (TypeError, ValueError):
                    # Skip items we can't properly inspect
                    continue

        except Exception as e:
            print(f"Warning: Error inspecting module {module.__name__}: {e}")

        return api

    def _explore_submodules(
        self, module: ModuleType, current_depth: int = 0
    ) -> dict[str, Any]:
        """Recursively explore submodules and their APIs.

        Args:
            module: The module to explore
            current_depth: The current depth of the exploration
        """
        result = {
            "api": self._get_module_public_api(module, module.__name__),
            "submodules": {},
        }

        if current_depth < self.max_depth:
            try:
                if hasattr(module, "__path__"):
                    for _, name, _ in pkgutil.iter_modules(module.__path__):
                        try:
                            full_name = f"{module.__name__}.{name}"
                            submodule = self._import_module(full_name)
                            if submodule:
                                result["submodules"][name] = self._explore_submodules(
                                    submodule, current_depth + 1
                                )
                        except Exception as e:
                            print(f"Warning: Error exploring submodule {name}: {e}")
                            continue
            except Exception as e:
                print(
                    f"Warning: Error accessing module path for {module.__name__}: {e}"
                )

        return result

    def explore(self) -> dict[str, Any]:
        """Explore the module tree starting from the root module.

        Returns:
            dict[str, Any]: The explored module tree
        """
        root_module = self._import_module(self.root_module_path)
        if root_module:
            self.tree = self._explore_submodules(root_module)
        return self.tree

    def get_tree_string(
        self, tree: dict[str, Any] | None = None, prefix: str = "", is_last: bool = True
    ) -> str:
        """Generate the module tree as a string in a hierarchical format.

        Args:
            tree: The module tree to generate a string for
            prefix: The prefix to use for the tree
            is_last: Whether the current module is the last in its parent
        """
        lines = []
        if tree is None:
            tree = self.tree
            lines.append(f"📦 {self.root_module_path}")

        api = tree.get("api", {})
        indent = "    " if is_last else "│   "

        # Add public API categories
        if api.get("all"):
            lines.append(
                f"{prefix}{'└── ' if is_last else '├── '}📜 __all__: {', '.join(api['all'])}"
            )

        for category, items in api.items():
            if category != "all" and items:
                lines.append(
                    f"{prefix}{'└── ' if is_last and not tree['submodules'] else '├── '}"
                    f"{'🔷' if category == 'classes' else '⚡' if category == 'functions' else '📌'} "
                    f"{category}: {', '.join(sorted(items))}"
                )

        # Add submodules
        submodules = tree.get("submodules", {})
        for idx, (name, subtree) in enumerate(submodules.items()):
            is_last_module = idx == len(submodules) - 1
            lines.append(f"{prefix}{'└── ' if is_last_module else '├── '}📦 {name}")
            lines.extend(
                self.get_tree_string(
                    subtree, prefix + indent, is_last_module
                ).splitlines()
            )

        return "\n".join(lines)


def display_signature(import_path: str) -> str:
    """
    Format a function's signature with clear visual organization.

    Args:
        import_path: Import path to the function (e.g., 'prefect.serve' or 'prefect.flows:serve')

    Returns:
        A formatted string showing the function signature

    Example:
        >>> print(display_signature('prefect.flows:serve'))
        📎 serve
        ├── Parameters:
        ├── flow: Flow
        └── name: str = None
        └── Returns:
            └── DeploymentResponse
    """
    try:
        func = import_object(import_path)
        if not callable(func):
            raise ValueError(f"Imported object {import_path!r} is not callable")
    except Exception as e:
        return f"Error: Could not import {import_path!r}: {str(e)}"

    sig = inspect.signature(func)

    # Get function name and build header
    func_name = func.__name__
    lines = [f"📎 {func_name}", "├── Parameters:"]

    # Process parameters
    for idx, (name, param) in enumerate(sig.parameters.items()):
        is_last = idx == len(sig.parameters) - 1
        prefix = "└── " if is_last else "├── "

        # Determine parameter kind
        if param.kind == inspect.Parameter.VAR_POSITIONAL:
            kind_str = "*args"
        elif param.kind == inspect.Parameter.VAR_KEYWORD:
            kind_str = "**kwargs"
        elif param.kind == inspect.Parameter.KEYWORD_ONLY:
            kind_str = "keyword-only"
        elif param.kind == inspect.Parameter.POSITIONAL_ONLY:
            kind_str = "positional-only"
        else:
            kind_str = ""

        # Format type annotation
        type_str = (
            ""
            if param.annotation == inspect.Parameter.empty
            else f": {param.annotation}"
        )

        # Format default value
        if param.default == inspect.Parameter.empty:
            default_str = ""
        else:
            default_str = f" = {repr(param.default)}"

        # Build the parameter line
        param_parts = []
        param_parts.append(name)
        if type_str:
            param_parts.append(type_str)
        if default_str:
            param_parts.append(default_str)
        if kind_str:
            param_parts.append(f" ({kind_str})")

        lines.append(f"{prefix}{''.join(param_parts)}")

    # Add return type if present
    return_annotation = sig.return_annotation
    if return_annotation != inspect.Signature.empty:
        lines.append("└── Returns:")
        lines.append(f"    └── {return_annotation}")

    return "\n".join(lines)



import os
import subprocess
from datetime import datetime
from typing import Optional

import httpx
import turbopuffer as tpuf
from modules import ModuleTreeExplorer
from prefect.blocks.system import Secret
from pydantic import BaseModel, Field, field_validator
from raggy.vectorstores.tpuf import multi_query_tpuf
from strings import slice_tokens

import marvin
from marvin.utilities.logging import get_logger


def verify_import_statements(import_statements: list[str]) -> str:
    """
    Verify that import statements are valid.

    Args:
        import_statements: A list of import statements to verify.

    Returns:
        A string with the results of the verification.

    Example:
        >>> verify_import_statements(["from prefect import flow"])
        "✅ from prefect import flow"

        >>> verify_import_statements(["from prefect import schleeb", "from prefect import flow"])
        "❌ from prefect import schleeb: No module named 'schleeb'\n✅ from prefect import flow"
    """
    print(f"Verifying import statements: {import_statements}")
    results: list[str] = []
    for import_statement in import_statements:
        try:
            result = subprocess.run(
                ["uv", "run", "--no-project", "python", "-c", import_statement],
                capture_output=True,
                text=True,
                check=False,
            )
            if result.returncode == 0:
                results.append(f"✅ {import_statement}")
            else:
                error_msg = result.stderr.strip().split("\n")[-1]
                results.append(f"❌ {import_statement}: {error_msg}")
        except Exception as e:
            results.append(f"❌ {import_statement}: {str(e)}")
    return "\n".join(results)


def review_top_level_prefect_api() -> str:
    """
    Review the available submodules and the top-level API of Prefect.
    """
    explorer = ModuleTreeExplorer("prefect", max_depth=0)
    explorer.explore()
    summary = explorer.get_tree_string()
    print(summary)
    return summary


def review_common_3x_gotchas() -> list[str]:
    """If needed, review common sources of confusion for Prefect 3.x users."""
    tips = [
        "from_source('https://github.com/<owner>/<repo>') has replaced the GitHub block in Prefect 3.x",
        ".map and .submit are always synchronous, even if the underlying function is asynchronous",
        "futures returned by .map can be resolved together, like integers = double.map(range(10)).result()",
        "futures must be resolved by passing them to another task, returning them or manually calling .result() or .wait()",
        "agents are replaced by workers in prefect 3.x, work pools replace the infra blocks from prefect.infrastructure",
        "prefect 3.x uses pydantic 2 and server data from prefect 2.x is not compatible with 3.x",
        "Deployment.build_from_flow() is removed in 3.x, use some_flow.from_source(...).deploy(...) instead.",
    ]
    print(tips)
    return tips


def explore_module_offerings(module_path: str, max_depth: int = 1) -> str:
    """
    Explore and return the public API tree of a specific module and its submodules as a string.

    Args:
        module_path: String representing the module path (e.g., 'prefect.runtime')
        max_depth: Maximum depth to explore in the module tree (default: 2)

    Returns:
        str: A formatted string representation of the module tree

    Example:
        >>> explore_module_tree('prefect.runtime', max_depth=0)
    """
    explorer = ModuleTreeExplorer(module_path, max_depth)
    explorer.explore()
    summary = explorer.get_tree_string()
    print(summary)
    return summary


def search_prefect_2x_docs(queries: list[str]) -> str:
    """Searches the Prefect documentation for the given queries.

    It is best to use more than one, short query to get the best results.

    For example, given a question like:
    "Is there a way to get the task_run_id for a task from a flow run?"

    You might use the following queries:
    - "retrieve task run id from flow run"
    - "retrieve run metadata dynamically"

    """
    if not tpuf.api_key:
        tpuf.api_key = Secret.load("tpuf-api-key", _sync=True).get()  # type: ignore

    print(f"Searching about {queries} in Prefect 2.x docs")

    return multi_query_tpuf(queries, namespace="prefect-2", n_results=5)


def search_prefect_3x_docs(queries: list[str]) -> str:
    """Searches the Prefect documentation for the given queries.

    It is best to use more than one, short query to get the best results.

    For example, given a question like:
    "Is there a way to get the task_run_id for a task from a flow run?"

    You might use the following queries:
    - "retrieve task run id from flow run"

    """
    if not tpuf.api_key:
        tpuf.api_key = Secret.load("tpuf-api-key", _sync=True).get()  # type: ignore

    print(f"Searching about {queries} in Prefect 3.x docs")

    return multi_query_tpuf(queries, namespace="prefect-3", n_results=5)


def search_controlflow_docs(queries: list[str]) -> str:
    """Searches the ControlFlow documentation for the given queries.

    ControlFlow is an agentic framework built on top of Prefect 3.x.

    It is best to use more than one, short query to get the best results.
    """
    return multi_query_tpuf(queries, namespace="controlflow", n_results=5)


def get_latest_prefect_release_notes() -> str:
    """Gets the latest Prefect release notes"""
    url = "https://api.github.com/repos/PrefectHQ/prefect/releases/latest"

    print(f"Getting latest Prefect release notes from {url}")

    with httpx.Client() as client:
        response = client.get(url)
        release_notes = response.json().get("body")
        return release_notes


async def get_token() -> str:
    try:
        from prefect.blocks.system import Secret

        return (await Secret.load(name="github-token")).get()  # type: ignore
    except (ImportError, ValueError) as exc:
        getattr(get_logger("marvin"), "debug_kv")(
            (
                "Prefect Secret for GitHub token not retrieved. "
                f"{exc.__class__.__name__}: {exc}"
                "red"
            ),
        )

    try:
        return getattr(marvin.settings, "github_token")
    except AttributeError:
        pass

    if token := os.environ.get("MARVIN_GITHUB_TOKEN", ""):
        return token

    raise RuntimeError("GitHub token not found")


class GitHubUser(BaseModel):
    """GitHub user."""

    login: Optional[str] = None


class GitHubComment(BaseModel):
    """GitHub comment."""

    body: str = Field(default="")
    user: GitHubUser = Field(default_factory=GitHubUser)


class GitHubLabel(BaseModel):
    """GitHub label."""

    name: str = Field(default="")


class GitHubIssue(BaseModel):
    """GitHub issue."""

    created_at: datetime = Field(...)
    html_url: str = Field(...)
    number: int = Field(...)
    title: str = Field(default="")
    body: str | None = Field(default="")
    labels: list[GitHubLabel] = Field(default_factory=GitHubLabel)
    user: GitHubUser = Field(default_factory=GitHubUser)

    @field_validator("body")
    def validate_body(cls, v: str) -> str:
        if not v:
            return ""
        return v


async def search_github_issues(
    query: str,
    repo: str = "prefecthq/prefect",
    n: int = 3,
    api_token: str | None = None,
) -> str:
    """
    Use the GitHub API to search for issues in a given repository. Do
    not alter the default value for `n` unless specifically requested by
    a user.

    For example, to search for open issues about AttributeErrors with the
    label "bug" in PrefectHQ/prefect:
        - repo: prefecthq/prefect
        - query: label:bug is:open AttributeError
    """
    headers = {"Accept": "application/vnd.github.v3+json"}

    headers["Authorization"] = f"Bearer {api_token or await get_token()}"

    async with httpx.AsyncClient() as client:
        response = await client.get(
            "https://api.github.com/search/issues",
            headers=headers,
            params={
                "q": query if "repo:" in query else f"repo:{repo} {query}",
                "order": "desc",
                "per_page": n,
            },
        )
        response.raise_for_status()

    issues_data = response.json()["items"]

    # enforce 1000 token limit per body
    for issue in issues_data:
        if not issue["body"]:
            continue
        issue["body"] = slice_tokens(issue["body"], 1000)

    issues = [GitHubIssue(**issue) for issue in issues_data]

    summary = "\n\n".join(
        f"{issue.title} ({issue.html_url}):\n{issue.body}" for issue in issues
    )
    if not summary.strip():
        return "No issues found."
    return summary



"""Thread management for conversations.

This module provides the Thread class for managing conversation context.
"""

import uuid
from contextvars import ContextVar
from dataclasses import dataclass, field
from datetime import datetime
from typing import Any, Optional, Sequence

from pydantic import TypeAdapter
from pydantic_ai.messages import UserContent
from pydantic_ai.usage import Usage
from sqlalchemy import select

from marvin.database import (
    DBLLMCall,
    DBLLMCallMessage,
    DBMessage,
    DBThread,
    get_async_session,
    utc_now,
)
from marvin.engine.llm import ModelRequest
from marvin.utilities.asyncio import run_sync

from .engine.llm import AgentMessage, PydanticAIMessage, SystemMessage, UserMessage

# Message serialization adapter
message_adapter: TypeAdapter[PydanticAIMessage | list[PydanticAIMessage]] = TypeAdapter(
    PydanticAIMessage | list[PydanticAIMessage]
)


@dataclass(kw_only=True)
class Message:
    id: uuid.UUID = field(default_factory=uuid.uuid4)
    thread_id: str = field(default=None)
    message: PydanticAIMessage
    created_at: datetime = field(default_factory=utc_now)


@dataclass(kw_only=True)
class LLMCallMessages:
    prompt: list[Message]
    completion: list[Message]


@dataclass(kw_only=True)
class LLMCall:
    """Represents an LLM call."""

    id: uuid.UUID
    thread_id: str
    usage: Usage
    timestamp: datetime

    def get_messages(self) -> LLMCallMessages:
        """Get the messages for this LLM call."""
        return run_sync(self.get_messages_async())

    async def get_messages_async(self) -> LLMCallMessages:
        """Get the messages for this LLM call."""
        async with get_async_session() as session:
            # Query the llm_call_messages table to get all messages associated with this LLM call

            # Get all message associations for this LLM call ordered by their sequence
            query = (
                select(DBLLMCallMessage, DBMessage)
                .join(
                    DBMessage,
                    DBLLMCallMessage.message_id == DBMessage.id,
                )
                .where(DBLLMCallMessage.llm_call_id == self.id)
                .order_by(DBLLMCallMessage.order)
            )

            result = await session.execute(query)

            # Map the database enum values to our new terminology
            # REQUEST -> prompt, RESPONSE -> completion
            messages_by_role = {"prompt": [], "completion": []}

            for llm_call_message, db_message in result:
                if llm_call_message.in_initial_prompt:
                    messages_by_role["prompt"].append(db_message.to_message())
                else:
                    messages_by_role["completion"].append(db_message.to_message())

            return LLMCallMessages(
                prompt=messages_by_role["prompt"],
                completion=messages_by_role["completion"],
            )


# Global context var for current thread
_current_thread: ContextVar[Optional["Thread"]] = ContextVar(
    "current_thread",
    default=None,
)

# Track the last thread globally
_last_thread: Optional["Thread"] = None


@dataclass
class Thread:
    """Main runtime object for managing conversation context."""

    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    parent_id: str | None = None
    _db_thread: bool = field(default=False, init=False, repr=False)
    _tokens: list[Any] = field(default_factory=list, init=False, repr=False)

    def __post_init__(self):
        if not isinstance(self.id, str):
            raise ValueError("Thread ID must be a string")

    async def _ensure_thread_exists(self) -> None:
        """Ensure thread exists in database."""
        if self._db_thread:
            return

        # First, ensure tables are created in case of missing initialization
        try:
            from marvin.database import ensure_db_tables_exist

            ensure_db_tables_exist()
        except Exception as e:
            from marvin.utilities.logging import get_logger

            logger = get_logger(__name__)
            logger.warning(f"Failed to ensure DB tables exist: {e}")

        # Now we can safely create the thread
        try:
            async with get_async_session() as session:
                db_thread = await session.get(DBThread, self.id)
                if db_thread is not None:
                    self._db_thread = True
                else:
                    await DBThread.create(
                        session=session,
                        id=self.id,
                        parent_thread_id=self.parent_id,
                    )
                    self._db_thread = True
        except Exception as e:
            from marvin.utilities.logging import get_logger

            logger = get_logger(__name__)
            logger.error(f"Failed to create thread in database: {e}")
            # Re-raise to fail fast rather than letting downstream code handle DB errors
            raise

    def add_messages(self, messages: list[PydanticAIMessage]) -> list[Message]:
        """Add multiple messages to the thread.

        Args:
            messages: List of messages to add (UserMessage, AssistantMessage, etc.)
            llm_call_id: Optional ID of the LLM call that generated these messages

        """
        return run_sync(self.add_messages_async(messages))

    async def add_messages_async(
        self, messages: list[PydanticAIMessage]
    ) -> list[Message]:
        """Add multiple messages to the thread.

        Args:
            messages: List of messages to add (UserMessage, AssistantMessage, etc.)
            llm_call_id: Optional ID of the LLM call that generated these messages

        """
        await self._ensure_thread_exists()

        async with get_async_session() as session:
            # Create DB message records
            db_messages = [
                DBMessage.from_message(thread_id=self.id, message=message)
                for message in messages
            ]
            session.add_all(db_messages)

            await session.commit()

        return [db_m.to_message() for db_m in db_messages]

    def add_system_message(self, message: str) -> Message:
        """Add a system message to the thread."""
        return run_sync(self.add_system_message_async(message))

    async def add_system_message_async(self, message: str) -> Message:
        """Add a system message to the thread."""
        messages = await self.add_messages_async([SystemMessage(content=message)])
        return messages[0]

    def add_user_message(self, message: str | Sequence[UserContent]) -> Message:
        """Add a user message to the thread."""
        return run_sync(self.add_user_message_async(message))

    async def add_user_message_async(
        self, message: str | Sequence[UserContent]
    ) -> Message:
        """Add a user message to the thread."""
        messages = await self.add_messages_async([UserMessage(content=message)])
        return messages[0]

    def add_agent_message(self, message: str) -> Message:
        """Add an agent message to the thread."""
        return run_sync(self.add_agent_message_async(message))

    async def add_agent_message_async(self, message: str) -> Message:
        """Add an agent message to the thread."""
        messages = await self.add_messages_async([AgentMessage(content=message)])
        return messages[0]

    def add_info_message(self, message: str, prefix: str = None) -> Message:
        """Add an info message to the thread."""
        return run_sync(self.add_info_message_async(message, prefix=prefix))

    async def add_info_message_async(self, message: str, prefix: str = None) -> Message:
        """Add an info message to the thread."""
        prefix = prefix or "INFO MESSAGE"
        messages = await self.add_messages_async(
            [UserMessage(content=f"{prefix}: {message}")]
        )
        return messages[0]

    def get_messages(
        self,
        before: datetime | None = None,
        after: datetime | None = None,
        limit: int | None = None,
        include_system_messages: bool = False,
    ) -> list[Message]:
        """Get all messages in this thread.

        Args:
            before: Only return messages before this timestamp
            after: Only return messages after this timestamp
            limit: Maximum number of messages to return
            include_system_messages: Whether to include system messages
        Returns:
            List of messages in chronological order
        """
        return run_sync(
            self.get_messages_async(
                before=before,
                after=after,
                limit=limit,
                include_system_messages=include_system_messages,
            ),
        )

    async def get_messages_async(
        self,
        before: datetime | None = None,
        after: datetime | None = None,
        limit: int | None = None,
        include_system_messages: bool = False,
    ) -> list[Message]:
        """Get all messages in this thread.

        Args:
            before: Only return messages before this timestamp
            after: Only return messages after this timestamp
            limit: Maximum number of messages to return
            include_system_messages: Whether to include system messages
        Returns:
            List of messages in chronological order
        """
        await self._ensure_thread_exists()

        async with get_async_session() as session:
            query = (
                select(DBMessage)
                .where(DBMessage.thread_id == self.id)
                .order_by(DBMessage.created_at.desc())
            )

            if before is not None:
                query = query.where(DBMessage.created_at < before)
            if after is not None:
                query = query.where(DBMessage.created_at > after)

            if limit is not None:
                query = query.limit(limit)

            result = await session.execute(query)

            db_messages = list(result.scalars().all())
            db_messages.reverse()

            messages = [db_m.to_message() for db_m in db_messages]

            if not include_system_messages:
                for m in messages:
                    if isinstance(m.message, ModelRequest) and any(
                        p.part_kind == "system-prompt" for p in m.message.parts
                    ):
                        messages.remove(m)

            return messages

    async def get_llm_calls_async(
        self,
        before: datetime | None = None,
        after: datetime | None = None,
        limit: int | None = None,
    ) -> list[LLMCall]:
        """Get LLM calls for this thread.

        Args:
            before: Only return calls before this timestamp
            after: Only return calls after this timestamp
            limit: Maximum number of calls to return

        Returns:
            List of LLM calls in chronological order
        """
        await self._ensure_thread_exists()

        async with get_async_session() as session:
            query = select(DBLLMCall).where(DBLLMCall.thread_id == self.id)

            if before is not None:
                query = query.where(DBLLMCall.timestamp < before)
            if after is not None:
                query = query.where(DBLLMCall.timestamp > after)

            query = query.order_by(DBLLMCall.timestamp)

            if limit is not None:
                query = query.limit(limit)

            result = await session.execute(query)
            return [
                LLMCall(
                    id=call.id,
                    thread_id=call.thread_id,
                    usage=call.usage,
                    timestamp=call.timestamp,
                )
                for call in result.scalars().all()
            ]

    def get_llm_calls(
        self,
        before: datetime | None = None,
        after: datetime | None = None,
        limit: int | None = None,
    ) -> list[LLMCall]:
        """Get LLM calls for this thread.

        Args:
            before: Only return calls before this timestamp
            after: Only return calls after this timestamp
            limit: Maximum number of calls to return

        Returns:
            List of LLM calls in chronological order
        """
        return run_sync(
            self.get_llm_calls_async(before=before, after=after, limit=limit)
        )

    async def get_usage_async(
        self,
        before: datetime | None = None,
        after: datetime | None = None,
    ) -> Usage:
        """Get the usage for this thread.

        Args:
            before: Only include usage before this timestamp
            after: Only include usage after this timestamp

        Returns:
            Total usage for the specified time range
        """
        await self._ensure_thread_exists()

        usage = Usage()

        llm_calls = await self.get_llm_calls_async(before=before, after=after)
        for llm_call in llm_calls:
            usage.incr(llm_call.usage)

        return usage

    def get_usage(
        self,
        before: datetime | None = None,
        after: datetime | None = None,
    ) -> Usage:
        """Get the usage for this thread.

        Args:
            before: Only include usage before this timestamp
            after: Only include usage after this timestamp

        Returns:
            Total usage for the specified time range
        """
        return run_sync(self.get_usage_async(before=before, after=after))

    def __enter__(self):
        """Set this thread as the current thread in context."""
        token = _current_thread.set(self)
        self._tokens.append(token)
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any):
        """Reset the current thread in context and store it as the last thread."""
        global _last_thread
        # Store this thread as the last thread before resetting current
        _last_thread = self
        if self._tokens:  # Only reset if we have tokens
            _current_thread.reset(self._tokens.pop())

    @classmethod
    def get_current(cls) -> Optional["Thread"]:
        """Get the current thread from context."""
        return _current_thread.get()


def get_current_thread() -> Thread | None:
    """Get the currently active thread from context.

    Returns:
        The current Thread instance or None if no thread is active.

    """
    return Thread.get_current()


def get_last_thread() -> Thread | None:
    """Get the last thread that was set as current.

    This function is intended for debugging purposes only, and will only work in
    certain contexts where the last thread is available in memory.
    """
    return _last_thread


def get_thread(thread: Thread | str | None) -> Thread:
    """Get a thread from the given input.

    Args:
        thread: Thread instance, thread ID, or None

    Returns:
        A Thread instance

    """
    if isinstance(thread, Thread):
        return thread
    elif thread is not None:
        return Thread(id=thread)
    else:
        return get_current_thread() or Thread()



from pathlib import Path

import marvin


def write_to_file(content: str, filename: str):
    Path(filename).write_text(content)


def read_file(filename: str) -> str:
    return Path(filename).read_text()


def delete_file(filename: str):
    Path(filename).unlink()


def confirm_with_user(content: str) -> bool:
    """require 'y' or 'yes' to confirm"""
    return input(content).lower() in ("y", "yes")


if __name__ == "__main__":
    agent = marvin.Agent(
        tools=[write_to_file, read_file, delete_file, confirm_with_user],
        prompt="use your tools to help the user with their request",
    )
    agent.run(
        (
            "write a file called 'test.txt' with content 'hello world',"
            "read the file"
            "and then delete the file if the user confirms this"
        ),
    )



import chromadb
import pytest
from qdrant_client import AsyncQdrantClient

import marvin
from marvin.memory.memory import MemoryProvider
from marvin.memory.providers.chroma import ChromaEphemeralMemory, ChromaMemory
from marvin.memory.providers.qdrant import QdrantMemory


@pytest.mark.parametrize(
    "provider",
    [
        ChromaEphemeralMemory(),
        QdrantMemory(client=AsyncQdrantClient(location=":memory:")),
    ],
)
class TestMemory:
    async def test_store_and_retrieve(self, provider: MemoryProvider):
        m = marvin.Memory(key="test", instructions="test", provider=provider)
        await m.add("The number is 42")
        result = await m.search("numbers")
        assert len(result) == 1
        assert "The number is 42" in result.values()

    async def test_delete(self, provider: MemoryProvider):
        m = marvin.Memory(key="test2", instructions="test", provider=provider)
        m_id = await m.add("The number is 42")
        await m.delete(m_id)
        result = await m.search("numbers")
        assert len(result) == 0

    async def test_search(self, provider: MemoryProvider):
        m = marvin.Memory(key="test3", instructions="test", provider=provider)
        await m.add("The number is 42")
        await m.add("The number is 43")
        result = await m.search("numbers")
        assert len(result) == 2
        assert "The number is 42" in result.values()
        assert "The number is 43" in result.values()


class TestMemoryProvider:
    def test_load_from_string_invalid(self):
        with pytest.raises(ValueError):
            marvin.Memory(key="test", instructions="test", provider="invalid")

    def test_load_from_string_chroma_db(self):
        m = marvin.Memory(key="test", instructions="test", provider="chroma-db")
        assert isinstance(m.provider, ChromaMemory)

    def test_load_from_instance(self, tmp_path):
        mp = ChromaMemory(
            client=chromadb.PersistentClient(path=str(tmp_path / "test_path")),
        )
        m = marvin.Memory(key="test", instructions="test", provider=mp)
        assert m



import inspect
from collections.abc import Callable
from dataclasses import dataclass
from functools import wraps
from typing import Any, Literal, TypeVar

import pydantic_ai
from pydantic_ai import RunContext

from marvin.utilities.logging import get_logger

T = TypeVar("T")
logger = get_logger(__name__)


def update_fn(
    func: Callable[..., T] | None = None,
    /,
    *,
    name: str | None = None,
    description: str | None = None,
) -> Callable[[Callable[..., T]], Callable[..., T]] | Callable[..., T]:
    """Update a function's name and optionally set its docstring.

    Can be used either as a decorator with keyword arguments or as a direct function.

    Args:
        func: The function to update (optional). If provided, updates are applied directly.
             If not provided, returns a decorator.
        name: The new name for the function (optional). If provided, must not be empty.
        description: Optional docstring for the function

    Example:
        # As a function:
        def my_fn(x):
            return x
        updated_fn = update_fn(my_fn, name='hello_there')

        # As a decorator with name:
        @update_fn(name='hello_there')
        def my_fn(x):
            return x

        # As a decorator with name and description:
        @update_fn(name='hello_there', description='Says hello')
        def my_fn(x):
            return x

        # As a decorator with no arguments:
        @update_fn()
        def my_fn(x):
            return x

        # Works with async functions too:
        @update_fn(name='async_hello')
        async def my_async_fn(x):
            return x
    """
    if name is not None and not name:
        raise ValueError("name cannot be empty if provided")

    def decorator(fn: Callable[..., T]) -> Callable[..., T]:
        if inspect.iscoroutinefunction(fn):

            @wraps(fn)
            async def wrapper(*args: Any, **kwargs: Any) -> T:
                return await fn(*args, **kwargs)
        else:

            @wraps(fn)
            def wrapper(*args: Any, **kwargs: Any) -> T:
                return fn(*args, **kwargs)

        if name is not None:
            wrapper.__name__ = name
        if description is not None:
            wrapper.__doc__ = description
        return wrapper

    # If func is provided, apply the decorator directly
    if func is not None:
        return decorator(func)

    # Otherwise return the decorator for use with @ syntax
    return decorator


@dataclass
class ResultTool:
    type: Literal["result-tool"] = "result-tool"

    def run(self, ctx: RunContext) -> None:
        pass


def wrap_tool_errors(tool_fn: Callable[..., Any]):
    """
    Pydantic AI doesn't catch errors except for ModelRetry, so we need to make
    sure we catch them ourselves and raise a ModelRetry instead.
    """
    if inspect.iscoroutinefunction(tool_fn):

        @wraps(tool_fn)
        async def _fn(*args, **kwargs):
            try:
                return await tool_fn(*args, **kwargs)
            except pydantic_ai.ModelRetry as e:
                logger.debug(f"Tool failed: {e}")
                raise e
            except Exception as e:
                logger.debug(f"Tool failed: {e}")
                raise pydantic_ai.ModelRetry(message=f"Tool failed: {e}") from e

        return _fn

    else:

        @wraps(tool_fn)
        def _fn(*args: Any, **kwargs: Any):
            try:
                return tool_fn(*args, **kwargs)
            except pydantic_ai.ModelRetry as e:
                logger.debug(f"Tool failed: {e}")
                raise e
            except Exception as e:
                logger.debug(f"Tool failed: {e}")
                raise pydantic_ai.ModelRetry(message=f"Tool failed: {e}") from e

        return _fn



import inspect
from typing import Any, TypeVar, cast

from pydantic import conlist

import marvin
from marvin.agents.agent import Agent
from marvin.handlers.handlers import AsyncHandler, Handler
from marvin.thread import Thread
from marvin.utilities.asyncio import run_sync
from marvin.utilities.jinja import jinja_env
from marvin.utilities.jsonschema import JSONSchema
from marvin.utilities.types import TargetType

T = TypeVar("T")


DEFAULT_PROMPT = """
You are an expert data generator that always creates high-quality, random
examples of a description or type. The data you produce is relied on for
testing, examples, demonstrations, and more. You use inference or deduction
whenever necessary to supply missing or omitted data. You will be given
instructions or a type format, as well as a number of entities to generate. 

Unless the user explicitly says otherwise, assume they are request a VARIED and
REALISTIC selection of useful outputs that meet their criteria. However, you
should prefer common responses to uncommon ones.

If the user provides additional instructions or a description, assume they are
looking for examples that satisfy the description. Do not provide more
information than the user requests. For example, if they ask for various
technologies, give their names but do not explain what each technology is.
"""


async def generate_async(
    target: TargetType[T] | None = None,
    n: int = 1,
    instructions: str | None = None,
    agent: Agent | None = None,
    thread: Thread | str | None = None,
    context: dict[str, Any] | None = None,
    handlers: list[Handler | AsyncHandler] | None = None,
    prompt: str | None = None,
) -> list[T]:
    """Generates examples of a specific type or matching a description asynchronously.

    This function uses a language model to generate high-quality, random examples
    that match the specified type or description. The examples are returned as a
    list.

    Args:
        target: The type of entities to generate.
        n: The number of examples to generate. Defaults to 1.
        instructions: Optional instructions describing what to generate. Used to
            provide specific guidance about what kinds of examples to create.
        agent: Optional custom agent to use for generation. If not provided,
            the default agent will be used.
        thread: Optional thread for maintaining conversation context. Can be
            either a Thread object or a string thread ID.
        context: Optional dictionary of additional context to include in the task.
        handlers: Optional list of handlers to use for the task.
        prompt: Optional prompt to use for the task. If not provided, the default
            prompt will be used.
    Returns:
        A list of n generated entities of type T.

    """
    _target = target

    if target is None:
        _target = str

    if _target is str and instructions is None:
        raise ValueError("Instructions are required when generating string values.")

    task_context = context or {}
    task_context["Number to generate"] = n

    prompt = prompt or DEFAULT_PROMPT
    if instructions:
        prompt += f"\n\nYou must follow these instructions for your generation:\n{instructions}"

    task = marvin.Task[list[target]](
        name="Generation Task",
        instructions=prompt,
        context=task_context,
        result_type=conlist(_target, min_length=n, max_length=n),
        agents=[agent] if agent else None,
    )

    return cast(list[T], await task.run_async(thread=thread, handlers=handlers))


def generate(
    target: TargetType[T] | None = None,
    n: int = 1,
    instructions: str | None = None,
    agent: Agent | None = None,
    thread: Thread | str | None = None,
    context: dict[str, Any] | None = None,
    handlers: list[Handler | AsyncHandler] | None = None,
    prompt: str | None = None,
) -> list[T]:
    """Generates examples of a specific type or matching a description.

    This function uses a language model to generate high-quality, random examples
    that match the specified type or description. The examples are returned as a
    list.

    Args:
        target: The type of entities to generate.
        n: The number of examples to generate. Defaults to 1.
        instructions: Optional instructions describing what to generate. Used to
            provide specific guidance about what kinds of examples to create.
        agent: Optional custom agent to use for generation. If not provided,
            the default agent will be used.
        thread: Optional thread for maintaining conversation context. Can be
            either a Thread object or a string thread ID.
        context: Optional dictionary of additional context to include in the task.
        handlers: Optional list of handlers to use for the task.
        prompt: Optional prompt to use for the task. If not provided, the default
            prompt will be used.
    Returns:
        A list of n generated entities of type T.

    """
    return run_sync(
        generate_async(
            target=target,
            n=n,
            instructions=instructions,
            agent=agent,
            thread=thread,
            context=context,
            handlers=handlers,
            prompt=prompt,
        ),
    )


async def generate_schema_async(
    instructions: str,
    base_schema: JSONSchema | None = None,
    agent: Agent | None = None,
    thread: Thread | str | None = None,
    context: dict[str, Any] | None = None,
    handlers: list[Handler | AsyncHandler] | None = None,
    prompt: str | None = None,
) -> JSONSchema:
    """Generates a JSON schema from a description."""

    prompt = prompt or inspect.cleandoc("""
        Your job is to generate JSON schemas that match the user's instructions. The latest instruction is:
        
        <instructions>
        {{instructions}}
        </instructions>
        
        Don't include extra nesting (e.g. const) unless the user explicitly asks for it - ie simple as possible.
        ---
        {% if base_schema %}
        Base your response on the following schema as much as possible:
        <base_schema>
        {{ base_schema }}
        </base_schema>
        {% else %}
        Base your response on previous instructions if possible.
        {% endif %}
        
        """)

    task = marvin.Task[JSONSchema](
        name="JSONSchema Generation",
        instructions=jinja_env.from_string(prompt).render(
            instructions=instructions,
            base_schema=base_schema,
        ),
        context=context,
        result_type=JSONSchema,
        agents=[agent] if agent else None,
    )

    return await task.run_async(thread=thread, handlers=handlers)


def generate_schema(
    instructions: str,
    base_schema: JSONSchema | None = None,
    agent: Agent | None = None,
    thread: Thread | str | None = None,
    context: dict[str, Any] | None = None,
    handlers: list[Handler | AsyncHandler] | None = None,
    prompt: str | None = None,
) -> JSONSchema:
    """Generates a JSON schema from a description."""
    return run_sync(
        generate_schema_async(
            instructions=instructions,
            base_schema=base_schema,
            agent=agent,
            thread=thread,
            context=context,
            handlers=handlers,
            prompt=prompt,
        )
    )



from marvin import Task


class TestTaskPrompt:
    async def test_task_return_string_prompt(self):
        # default tasks return a list; ensure strings work as well
        class MyTask(Task):
            def get_prompt(self) -> str:
                return 'Say the word "hello"'

        task = MyTask("")
        prompt = task.get_prompt()
        assert prompt == 'Say the word "hello"'
        assert task.run() == "hello"



"""Agents for Marvin.

An Agent is an entity that can process tasks and maintain state across interactions.
"""

import inspect
import random
from collections.abc import Callable
from dataclasses import dataclass, field
from pathlib import Path
from typing import TYPE_CHECKING, Any, TypeVar, Union

import pydantic_ai
from pydantic_ai.models import KnownModelName, Model, ModelSettings
from pydantic_ai.result import ToolOutput

import marvin
from marvin.agents.actor import Actor
from marvin.agents.names import AGENT_NAMES
from marvin.memory.memory import Memory
from marvin.prompts import Template
from marvin.utilities.logging import get_logger
from marvin.utilities.tools import wrap_tool_errors
from marvin.utilities.types import issubclass_safe

logger = get_logger(__name__)
T = TypeVar("T")

if TYPE_CHECKING:
    from marvin.engine.end_turn import EndTurn
    from marvin.engine.events import Event
    from marvin.handlers.handlers import AsyncHandler, Handler


async def handle_event(
    event: "Event", handlers: list["Handler | AsyncHandler"] | None = None
):
    """Handle an event by passing it to all registered handlers."""
    for handler in handlers or []:
        if isinstance(handler, AsyncHandler):
            await handler._handle(event)
        else:
            handler._handle(event)


@dataclass(kw_only=True)
class Agent(Actor):
    """An agent that can process tasks and maintain state."""

    name: str = field(
        default_factory=lambda: random.choice(AGENT_NAMES),
        metadata={"description": "Name of the agent"},
        kw_only=False,
    )

    tools: list[Callable[..., Any]] = field(
        default_factory=lambda: [],
        metadata={"description": "List of tools available to the agent"},
    )

    memories: list[Memory] = field(
        default_factory=lambda: [],
        metadata={"description": "List of memory modules available to the agent"},
    )

    model: KnownModelName | Model | None = field(
        default=None,
        metadata={
            "description": inspect.cleandoc("""
                The model to use for the agent. If not provided,
                the default model will be used. A compatible string
                can be passed to automatically retrieve the model.
                """),
        },
        repr=False,
    )

    model_settings: ModelSettings = field(
        default_factory=ModelSettings,
        metadata={"description": "Settings to pass to the model"},
        repr=False,
    )

    prompt: str | Path = field(
        default=Path("agent.jinja"),
        metadata={"description": "Template for the agent's prompt"},
        repr=False,
    )

    def __hash__(self) -> int:
        return super().__hash__()

    def get_model(self) -> Model | KnownModelName:
        return self.model or marvin.defaults.model

    def get_tools(self) -> list[Callable[..., Any]]:
        tools = self.tools + [t for m in self.memories for t in m.get_tools()]
        return tools

    def get_memories(self) -> list[Memory]:
        return list(self.memories)

    def get_model_settings(self) -> ModelSettings:
        defaults: ModelSettings = {}
        if marvin.settings.agent_temperature is not None:
            defaults["temperature"] = marvin.settings.agent_temperature
        return defaults | self.model_settings

    def _determine_result_type(self, end_turn_tools: list[Any]) -> type:
        # Simplified logic: if exactly one end-turn tool, use its type,
        # otherwise return a Union of all available end-turn types.
        if len(end_turn_tools) == 1:
            return end_turn_tools[0]
        else:
            from typing import Union

            return Union[tuple(end_turn_tools)]

    async def get_agentlet(
        self,
        tools: list[Callable[..., Any]],
        end_turn_tools: list["EndTurn"],
    ) -> pydantic_ai.Agent[Any, Any]:
        from marvin.engine.end_turn import EndTurn

        tools = tools + self.get_tools()
        end_turn_tools = end_turn_tools + self.get_end_turn_tools()

        # if any tools are actually EndTurn classes, remove them from tools and
        # add them to end turn tools
        for t in tools:
            if issubclass_safe(t, EndTurn):
                tools.remove(t)
                end_turn_tools.append(t)

        if not end_turn_tools:
            result_type = [EndTurn]

        if len(end_turn_tools) == 1:
            result_type = end_turn_tools[0]
            result_tool_name = result_type.__name__
        else:
            result_type = Union[tuple(end_turn_tools)]
            result_tool_name = "EndTurn"

        tools = [wrap_tool_errors(tool) for tool in tools or []]

        agentlet = pydantic_ai.Agent[Any, result_type](  # type: ignore
            model=self.get_model(),
            output_type=ToolOutput[result_type](
                type_=result_type,
                name=result_tool_name or "EndTurn",
                description="This tool will end your turn. You may only use one turn-ending tool per turn.",
            ),
            tools=tools,
            model_settings=self.get_model_settings(),
            end_strategy="exhaustive",
            retries=marvin.settings.agent_retries,
        )
        # new fields
        agentlet._marvin_tools = tools
        agentlet._marvin_end_turn_tools = end_turn_tools
        return agentlet

    def get_prompt(self) -> str:
        return Template(source=self.prompt).render(agent=self)



import datetime
import uuid
from dataclasses import dataclass, field
from typing import Any, Callable, Literal

import partial_json_parser
from pydantic_ai.messages import (
    FinalResultEvent,
    ModelResponsePart,
    RetryPromptPart,
    TextPart,
    TextPartDelta,
    ToolCallPart,
    ToolCallPartDelta,
    ToolReturnPart,
    UserPromptPart,
)
from pydantic_ai.result import FinalResult

from marvin.agents.actor import Actor
from marvin.engine.end_turn import EndTurn

# Define event types as literals for type checking
EventType = Literal[
    "user-message",
    "tool-result",
    "tool-retry",
    "tool-call",
    "tool-call-delta",
    "actor-message",
    "actor-message-delta",
    "orchestrator-start",
    "orchestrator-end",
    "orchestrator-error",
    "actor-start-turn",
    "actor-end-turn",
    "end-turn-tool-call",
    "end-turn-tool-result",
]


@dataclass(kw_only=True)
class Event:
    """Base class for all events in the system."""

    _dataclass_config = dict(kw_only=True)

    type: EventType
    id: uuid.UUID = field(default_factory=lambda: uuid.uuid4())
    timestamp: datetime.datetime = field(
        default_factory=lambda: datetime.datetime.now(datetime.timezone.utc),
    )


@dataclass(kw_only=True)
class UserMessageEvent(Event):
    """Event for user messages."""

    type: EventType = field(default="user-message", init=False)
    message: UserPromptPart


@dataclass(kw_only=True)
class ToolResultEvent(Event):
    """Event for tool return values."""

    type: EventType = field(default="tool-result", init=False)
    message: ToolReturnPart


@dataclass(kw_only=True)
class ToolRetryEvent(Event):
    """Event for tool retry requests."""

    type: EventType = field(default="tool-retry", init=False)
    message: RetryPromptPart


@dataclass(kw_only=True)
class ToolCallEvent(Event):
    """Event for complete tool calls ready to be executed."""

    type: EventType = field(default="tool-call", init=False)
    actor: Actor
    message: ToolCallPart
    tool_call_id: str
    tool: Callable[..., Any] | None

    def args_dict(self) -> dict[str, Any]:
        """Return the args as a dictionary."""
        if self.message.args and isinstance(self.message.args, str):
            return partial_json_parser.loads(self.message.args)
        return self.message.args


@dataclass(kw_only=True)
class EndTurnToolCallEvent(Event):
    """Event that fires as soon as we know that an end turn tool call has been made."""

    type: EventType = field(default="end-turn-tool-call", init=False)
    actor: Actor
    event: FinalResultEvent
    tool_call_id: str
    tool: EndTurn


@dataclass(kw_only=True)
class EndTurnToolResultEvent(Event):
    """Event for the final result from an end turn tool."""

    type: EventType = field(default="end-turn-tool-result", init=False)
    actor: Actor
    result: FinalResult
    tool_call_id: str
    tool: EndTurn


@dataclass(kw_only=True)
class ToolCallDeltaEvent(Event):
    """Event for delta updates to tool calls during streaming."""

    type: EventType = field(default="tool-call-delta", init=False)
    actor: Actor
    delta: ToolCallPartDelta
    snapshot: ToolCallPart
    tool_call_id: str
    tool: Callable[..., Any] | None

    def args_dict(self) -> dict[str, Any]:
        """Return the args as a dictionary."""
        if self.snapshot.args and isinstance(self.snapshot.args, str):
            return partial_json_parser.loads(self.snapshot.args)
        return self.snapshot.args


@dataclass(kw_only=True)
class ActorMessageEvent(Event):
    """Event for complete text messages from an agent."""

    type: EventType = field(default="actor-message", init=False)
    actor: Actor
    message: TextPart


@dataclass(kw_only=True)
class ActorMessageDeltaEvent(Event):
    """Event for delta updates to agent messages during streaming."""

    type: EventType = field(default="actor-message-delta", init=False)
    actor: Actor
    delta: TextPartDelta
    snapshot: TextPart


@dataclass(kw_only=True)
class OrchestratorStartEvent(Event):
    """Event for orchestrator start."""

    type: EventType = field(default="orchestrator-start", init=False)


@dataclass(kw_only=True)
class OrchestratorEndEvent(Event):
    """Event for orchestrator end."""

    type: EventType = field(default="orchestrator-end", init=False)


@dataclass(kw_only=True)
class OrchestratorErrorEvent(Event):
    """Event for orchestrator exceptions."""

    type: EventType = field(default="orchestrator-error", init=False)
    error: str


@dataclass(kw_only=True)
class ActorStartTurnEvent(Event):
    """Event for agent turn start."""

    type: EventType = field(default="actor-start-turn", init=False)
    actor: Actor


@dataclass(kw_only=True)
class ActorEndTurnEvent(Event):
    """Event for agent turn end."""

    type: EventType = field(default="actor-end-turn", init=False)
    actor: Actor


# Helper function to extract text from parts
def get_text_from_parts(parts: list[ModelResponsePart]) -> str:
    """Extract text content from a list of ModelResponseParts."""
    text_parts = []
    for part in parts:
        if isinstance(part, TextPart):
            text_parts.append(part.content)
    return "".join(text_parts)



"""Test configuration and fixtures."""

import threading
from pathlib import Path
from tempfile import TemporaryDirectory

import chromadb
import pytest
from pydantic_ai.models.test import TestModel

import marvin
from marvin import database, settings
from marvin.defaults import override_defaults
from marvin.instructions import instructions
from marvin.memory.providers.chroma import ChromaMemory

# Lock for database operations
_db_lock = threading.Lock()


@pytest.fixture(autouse=True)
async def setup_test_db(monkeypatch: pytest.MonkeyPatch, worker_id: str):
    """Use a temporary database for tests.

    The worker_id fixture is provided by pytest-xdist and will be 'gw0', 'gw1', etc
    for parallel test runners, or 'master' for single-process runs.
    """
    with TemporaryDirectory() as temp_dir:
        original_path = settings.database_url

        # Create unique path per worker to avoid conflicts
        worker_suffix = worker_id if worker_id != "master" else ""
        temp_path = Path(temp_dir) / f"test{worker_suffix}.db"
        database_url = f"sqlite+aiosqlite:///{temp_path}"

        with _db_lock:
            # Configure database settings
            monkeypatch.setattr(settings, "database_url", database_url)

            database._async_engine_cache.clear()

            # Create tables with retries
            max_retries = 3
            for attempt in range(max_retries):
                try:
                    await database.create_db_and_tables(force=True)
                    break
                except Exception:
                    if attempt == max_retries - 1:
                        raise
                    continue

        yield

        settings.database_url = original_path
        # Clear engine cache
        database._async_engine_cache.clear()


@pytest.fixture
async def session():
    """Provide an async database session for tests."""
    async with database.get_async_session() as session:
        yield session


@pytest.fixture(autouse=True)
def setup_memory(tmp_path: Path, monkeypatch: pytest.MonkeyPatch, worker_id: str):
    monkeypatch.setattr(
        marvin.defaults,
        "memory_provider",
        ChromaMemory(
            client=chromadb.PersistentClient(
                path=str(tmp_path / "controlflow-memory" / worker_id)
            ),
        ),
    )


@pytest.fixture
def gpt_4o():
    with override_defaults(model="openai:gpt-4o"):
        yield


@pytest.fixture
def gpt_4o_audio_preview():
    with override_defaults(model="openai:gpt-4o-audio-preview"):
        yield


@pytest.fixture(autouse=True)
def restore_defaults():
    """
    Ensure that the defaults are restored after each test, in case the test
    modified them.
    """
    with override_defaults(**marvin.defaults.__dict__):
        yield


@pytest.fixture
def unit_test_instructions():
    with instructions(
        """
        You are being unit tested. Be as fast and concise as possible. Do not
        post unecessary messages.
        """
    ):
        yield


@pytest.fixture()
def test_model():
    model = TestModel()
    with override_defaults(model=model):
        yield model



import enum
from collections.abc import Sequence
from typing import Any, Literal, TypeVar, overload

import marvin
from marvin.agents.agent import Agent
from marvin.handlers.handlers import AsyncHandler, Handler
from marvin.thread import Thread
from marvin.utilities.asyncio import run_sync
from marvin.utilities.types import Labels, issubclass_safe

T = TypeVar("T")

DEFAULT_PROMPT = """
You are an expert classifier that always maintains as much semantic meaning
as possible when labeling text. You use inference or deduction whenever
necessary to understand missing or omitted data. Classify the provided `data`,
text, or information as one of the provided labels. For boolean labels,
consider "truthy" or affirmative inputs to be "true"."""

PROMPT = DEFAULT_PROMPT  # for backwards compatibility


@overload
async def classify_async(
    data: Any,
    labels: Sequence[T] | type[T],
    multi_label: Literal[False] = False,
    *,
    instructions: str | None = None,
    agent: Agent | None = None,
    thread: Thread | str | None = None,
    context: dict[str, Any] | None = None,
    handlers: list[Handler | AsyncHandler] | None = None,
    prompt: str | None = None,
) -> T: ...


@overload
async def classify_async(
    data: Any,
    labels: Sequence[T] | type[T],
    multi_label: Literal[True],
    *,
    instructions: str | None = None,
    agent: Agent | None = None,
    thread: Thread | str | None = None,
    context: dict[str, Any] | None = None,
    handlers: list[Handler | AsyncHandler] | None = None,
    prompt: str | None = None,
) -> list[T]: ...


@overload
async def classify_async(
    data: Any,
    labels: Sequence[T] | type[T],
    multi_label: bool = False,
    *,
    instructions: str | None = None,
    agent: Agent | None = None,
    thread: Thread | str | None = None,
    context: dict[str, Any] | None = None,
    handlers: list[Handler | AsyncHandler] | None = None,
    prompt: str | None = None,
) -> T | list[T]: ...


async def classify_async(
    data: Any,
    labels: Sequence[T] | type[T],
    multi_label: bool = False,
    *,
    instructions: str | None = None,
    agent: Agent | None = None,
    thread: Thread | str | None = None,
    context: dict[str, Any] | None = None,
    handlers: list[Handler | AsyncHandler] | None = None,
    prompt: str | None = None,
) -> T | list[T]:
    """Asynchronously classifies input data into one or more predefined labels using a language model.

    This function uses a language model to analyze the input data and assign it to
    the most appropriate label(s) from the provided sequence of labels or Enum class.

    Args:
        data: The input data to classify. Can be any type.
        labels: Either a sequence of possible labels (of type T) or an Enum class to
            classify the data into. If an Enum class is provided, its values will be
            used as the labels.
        multi_label: If False (default), returns a single label. If True, returns
            multiple labels as a list.
        instructions: Optional additional instructions to guide the classification.
            Used to provide specific guidance about how to interpret or classify
            the data.
        agent: Optional custom agent to use for classification. If not provided,
            the default agent will be used.
        thread: Optional thread for maintaining conversation context. Can be
            either a Thread object or a string thread ID.
        context: Optional dictionary of additional context to include in the task.
        prompt: Optional prompt to use for the task. If not provided, the default
            prompt will be used.
    Returns:
        - If labels is a Sequence[T]:
            - If multi_label is False: returns T
            - If multi_label is True: returns list[T]
        - If labels is an Enum class:
            - If multi_label is False: returns E (the Enum value)
            - If multi_label is True: returns list[E] (list of Enum values)

    Examples:
        >>> # Using a sequence of labels
        >>> await classify_async("red car", ["red", "blue", "green"])
        'red'

        >>> # Using an Enum class
        >>> class Colors(enum.Enum):
        ...     RED = "red"
        ...     BLUE = "blue"
        ...     GREEN = "green"
        >>> await classify_async("red car", Colors)
        <Colors.RED: 'red'>

        >>> # Multi-label classification
        >>> await classify_async("red and blue car", Colors, multi_label=True)
        [<Colors.RED: 'red'>, <Colors.BLUE: 'blue'>]

        >>> # Boolean classification
        >>> await classify_async("2+2=4", bool)
        True

    """
    task_context = context or {}
    task_context.update({"Data to classify": data})

    prompt = prompt or PROMPT
    if instructions:
        prompt += f"\n\nYou must follow these instructions for your classification:\n{instructions}"

    # Handle bool/enum types specially for correct typing
    if labels is bool or issubclass_safe(labels, enum.Enum):
        # For bool/enum, we need list[labels] for multi-label
        result_type = list[labels] if multi_label else labels  # Runtime type
        ReturnType = list[T] if multi_label else T  # Generic type
    else:
        # For sequences, we use Labels for runtime validation
        result_type = Labels(labels, many=multi_label)  # Runtime type
        ReturnType = list[T] if multi_label else T  # Generic type

    task = marvin.Task[ReturnType](
        name="Classification Task",
        instructions=prompt,
        context=task_context,
        result_type=result_type,
        agents=[agent] if agent else None,
    )

    return await task.run_async(thread=thread, handlers=handlers)  # type: ignore


@overload
def classify(
    data: Any,
    labels: Sequence[T] | type[T],
    multi_label: Literal[False] = False,
    *,
    instructions: str | None = None,
    agent: Agent | None = None,
    thread: Thread | str | None = None,
    context: dict[str, Any] | None = None,
    handlers: list[Handler | AsyncHandler] | None = None,
) -> T: ...


@overload
def classify(
    data: Any,
    labels: Sequence[T] | type[T],
    multi_label: Literal[True],
    *,
    instructions: str | None = None,
    agent: Agent | None = None,
    thread: Thread | str | None = None,
    context: dict[str, Any] | None = None,
    handlers: list[Handler | AsyncHandler] | None = None,
) -> list[T]: ...


@overload
def classify(
    data: Any,
    labels: Sequence[T] | type[T],
    multi_label: bool = False,
    *,
    instructions: str | None = None,
    agent: Agent | None = None,
    thread: Thread | str | None = None,
    context: dict[str, Any] | None = None,
    handlers: list[Handler | AsyncHandler] | None = None,
) -> T | list[T]: ...


def classify(
    data: Any,
    labels: Sequence[T] | type[T],
    multi_label: bool = False,
    *,
    instructions: str | None = None,
    agent: Agent | None = None,
    thread: Thread | str | None = None,
    context: dict[str, Any] | None = None,
    handlers: list[Handler | AsyncHandler] | None = None,
    prompt: str | None = None,
) -> T | list[T]:
    """Classifies input data into one or more predefined labels using a language model.

    This function uses a language model to analyze the input data and assign it to
    the most appropriate label(s) from the provided sequence of labels or Enum class.

    Args:
        data: The input data to classify. Can be any type.
        labels: Either a sequence of possible labels (of type T) or an Enum class to
            classify the data into. If an Enum class is provided, its values will be
            used as the labels.
        multi_label: If False (default), returns a single label. If True, returns
            multiple labels as a list.
        instructions: Optional additional instructions to guide the classification.
            Used to provide specific guidance about how to interpret or classify
            the data.
        agent: Optional custom agent to use for classification. If not provided,
            the default agent will be used.
        thread: Optional thread for maintaining conversation context. Can be
            either a Thread object or a string thread ID.
        context: Optional dictionary of additional context to include in the task.
        handlers: Optional list of handlers to use for the task.
    Returns:
        - If labels is a Sequence[T]:
            - If multi_label is False: returns T
            - If multi_label is True: returns list[T]
        - If labels is an Enum class:
            - If multi_label is False: returns E (the Enum value)
            - If multi_label is True: returns list[E] (list of Enum values)

    Examples:
        >>> # Using a sequence of labels
        >>> classify("red car", ["red", "blue", "green"])
        'red'

        >>> # Using an Enum class
        >>> class Colors(enum.Enum):
        ...     RED = "red"
        ...     BLUE = "blue"
        ...     GREEN = "green"
        >>> classify("red car", Colors)
        <Colors.RED: 'red'>

        >>> # Multi-label classification
        >>> classify("red and blue car", Colors, multi_label=True)
        [<Colors.RED: 'red'>, <Colors.BLUE: 'blue'>]

        >>> # Boolean classification
        >>> classify("2+2=4", bool)
        True

    """
    return run_sync(
        classify_async(
            data=data,
            labels=labels,
            multi_label=multi_label,
            instructions=instructions,
            agent=agent,
            thread=thread,
            context=context,
            handlers=handlers,
            prompt=prompt,
        ),
    )



import os
import subprocess
from pathlib import Path

import typer

dev_app = typer.Typer(no_args_is_help=True)


@dev_app.command()
def docs():
    """Run the Marvin documentation locally. Equivalent to 'cd docs && mintlify dev' from the Marvin root."""
    try:
        # Get the absolute path of the Marvin main repo
        repo_root = Path(__file__).resolve().parents[3]
        docs_path = repo_root / "docs"

        if not docs_path.exists():
            typer.echo(f"Error: Docs directory not found at {docs_path}", err=True)
            raise typer.Exit(code=1)

        typer.echo(f"Changing directory to: {docs_path}")
        os.chdir(docs_path)

        typer.echo("Running 'mintlify dev'...")
        subprocess.run(["mintlify", "dev"], check=True)

    except subprocess.CalledProcessError as e:
        typer.echo(f"Error running 'mintlify dev': {e!s}", err=True)
        raise typer.Exit(code=1)
    except Exception as e:
        typer.echo(f"An error occurred: {e!s}", err=True)
        raise typer.Exit(code=1)



import uuid
from dataclasses import dataclass, field
from typing import Any

import marvin
from marvin.memory.memory import MemoryProvider

try:
    import chromadb  # noqa: F401
except ImportError:
    raise ImportError(
        "To use Chroma as a memory provider, please install the `chromadb` package."
    )


@dataclass(kw_only=True)
class ChromaMemory(MemoryProvider):
    model_config = dict(arbitrary_types_allowed=True)
    client: chromadb.ClientAPI = field(
        default_factory=lambda: chromadb.PersistentClient(
            path=str(marvin.settings.home_path / "memory/chroma"),
        ),
    )
    collection_name: str = field(
        default="memory-{key}",
        metadata={
            "description": """
            Optional; the name of the collection to use. This should be a 
            string optionally formatted with the variable `key`, which 
            will be provided by the memory module. The default is `"memory-{key}"`.
            """,
        },
    )

    def get_collection(self, memory_key: str) -> chromadb.Collection:
        return self.client.get_or_create_collection(
            self.collection_name.format(key=memory_key),
        )

    async def add(self, memory_key: str, content: str) -> str:
        collection = self.get_collection(memory_key)
        memory_id = str(uuid.uuid4())
        collection.add(
            documents=[content],
            metadatas=[{"id": memory_id}],
            ids=[memory_id],
        )
        return memory_id

    async def delete(self, memory_key: str, memory_id: str) -> None:
        collection = self.get_collection(memory_key)
        collection.delete(ids=[memory_id])

    async def search(self, memory_key: str, query: str, n: int = 20) -> dict[str, str]:
        results = self.get_collection(memory_key).query(
            query_texts=[query],
            n_results=n,
        )
        assert results["ids"] and results["documents"], "No results found"
        return dict(zip(results["ids"][0], results["documents"][0], strict=False))


def ChromaEphemeralMemory(**kwargs: Any) -> ChromaMemory:
    return ChromaMemory(client=chromadb.EphemeralClient(**kwargs))


def ChromaPersistentMemory(path: str | None = None, **kwargs: Any) -> ChromaMemory:
    return ChromaMemory(
        client=chromadb.PersistentClient(
            path=path or str(marvin.settings.home_path / "memory" / "chroma"),
            **kwargs,
        ),
    )


def ChromaCloudMemory(
    tenant: str | None = None,
    database: str | None = None,
    api_key: str | None = None,
    **kwargs: Any,
) -> ChromaMemory:
    tenant = tenant or marvin.settings.chroma_cloud_tenant
    assert tenant, "Tenant is required"
    database = database or marvin.settings.chroma_cloud_database
    assert database, "Database is required"
    return ChromaMemory(
        client=chromadb.CloudClient(
            api_key=api_key or marvin.settings.chroma_cloud_api_key,
            tenant=tenant,
            database=database,
            **kwargs,
        ),
    )



import asyncio
from collections.abc import Callable, Sequence
from typing import Annotated, Any, AsyncGenerator, TypeVar

from pydantic_ai.messages import UserContent

import marvin
import marvin.utilities.asyncio
from marvin import Task, Thread
from marvin.agents.actor import Actor
from marvin.engine.events import Event
from marvin.engine.orchestrator import Orchestrator
from marvin.handlers.handlers import AsyncHandler, Handler

T = TypeVar("T")


async def run_tasks_async(
    tasks: list[Task[Any]],
    thread: Thread | str | None = None,
    raise_on_failure: bool = True,
    handlers: list[Handler | AsyncHandler] | None = None,
) -> list[Task[Any]] | AsyncGenerator[Event, None]:
    orchestrator = Orchestrator(
        tasks=tasks,
        thread=thread,
        handlers=handlers,
    )
    await orchestrator.run(raise_on_failure=raise_on_failure)
    return tasks


async def run_tasks_stream(
    tasks: list[Task[Any]],
    thread: Thread | str | None = None,
    raise_on_failure: bool = True,
    handlers: list[Handler | AsyncHandler] | None = None,
) -> AsyncGenerator[Event, None]:
    # Create and configure the queue handler.
    queue_handler = marvin.handlers.QueueHandler()
    handlers = (handlers or []) + [queue_handler]

    # Initialize the orchestrator with the handlers.
    orchestrator = Orchestrator(
        tasks=tasks,
        thread=thread,
        handlers=handlers,
    )

    # Start the orchestrator in the background.
    orchestrator_task = asyncio.create_task(
        orchestrator.run(raise_on_failure=raise_on_failure)
    )

    try:
        # Continue looping until the orchestrator finishes.
        while not orchestrator_task.done():
            try:
                # Wait for the next event from the queue.
                event = await asyncio.wait_for(queue_handler.queue.get(), timeout=0.1)
                yield event
            except asyncio.TimeoutError:
                # No event was available in the queue; loop again.
                continue

        # Once the orchestrator has finished, yield any remaining events.
        while not queue_handler.queue.empty():
            yield queue_handler.queue.get_nowait()

    finally:
        # Optionally, cancel the orchestrator if needed.
        if not orchestrator_task.done():
            orchestrator_task.cancel()


def run_tasks(
    tasks: list[Task[Any]],
    thread: Thread | str | None = None,
    raise_on_failure: bool = True,
    handlers: list[Handler | AsyncHandler] | None = None,
) -> list[Task[Any]]:
    return marvin.utilities.asyncio.run_sync(
        run_tasks_async(
            tasks=tasks,
            thread=thread,
            raise_on_failure=raise_on_failure,
            handlers=handlers,
        ),
    )


async def run_async(
    instructions: str | Sequence[UserContent],
    result_type: type[T] | Annotated[T, Any] = str,
    tools: list[Callable[..., Any]] = [],
    thread: Thread | str | None = None,
    agents: list[Actor] | None = None,
    handlers: list[Handler | AsyncHandler] | None = None,
    raise_on_failure: bool = True,
    **kwargs: Any,
) -> T | AsyncGenerator[Event, None]:
    task = Task[result_type](
        instructions=instructions,
        result_type=result_type,
        agents=agents,
        tools=tools,
        **kwargs,
    )
    await run_tasks_async(
        [task],
        thread=thread,
        raise_on_failure=raise_on_failure,
        handlers=handlers,
    )

    return task.result


def run(
    instructions: str | Sequence[UserContent],
    result_type: type[T] = str,
    tools: list[Callable[..., Any]] = [],
    thread: Thread | str | None = None,
    agents: list[Actor] | None = None,
    raise_on_failure: bool = True,
    handlers: list[Handler | AsyncHandler] | None = None,
    **kwargs: Any,
) -> T:
    return marvin.utilities.asyncio.run_sync(
        run_async(
            instructions=instructions,
            result_type=result_type,
            tools=tools,
            thread=thread,
            agents=agents,
            raise_on_failure=raise_on_failure,
            handlers=handlers,
            **kwargs,
        ),
    )


def run_stream(
    instructions: str | Sequence[UserContent],
    result_type: type[T] = str,
    tools: list[Callable[..., Any]] = [],
    thread: Thread | str | None = None,
    agents: list[Actor] | None = None,
    raise_on_failure: bool = True,
    handlers: list[Handler | AsyncHandler] | None = None,
    **kwargs: Any,
) -> AsyncGenerator[Event, None]:
    task = Task[result_type](
        instructions=instructions,
        result_type=result_type,
        agents=agents,
        tools=tools,
        **kwargs,
    )
    return run_tasks_stream(
        [task],
        thread=thread,
        raise_on_failure=raise_on_failure,
        handlers=handlers,
    )



from sqlalchemy import select
from sqlalchemy.orm import selectinload

from marvin.database import (
    DBMessage,
    DBThread,
    create_db_and_tables,
)


async def test_async_session(session):
    """Test that async sessions are properly created and closed."""
    # Do some operation to ensure session works
    result = await session.execute(select(DBThread))
    result.all()  # Explicitly consume the result
    assert session.is_active


async def test_force_recreate_tables(session):
    """Test that tables can be force recreated."""
    # First phase: Create data
    thread = DBThread(id="test-thread")
    session.add(thread)
    await session.commit()
    await session.refresh(thread)

    message = DBMessage(
        thread_id=thread.id,
        message={"role": "user", "content": "test", "kind": "request"},
    )
    session.add(message)
    await session.commit()

    # Second phase: Verify data exists
    result = await session.execute(select(DBThread).where(DBThread.id == "test-thread"))
    assert result.scalars().first() is not None

    result = await session.execute(
        select(DBMessage).where(DBMessage.thread_id == "test-thread"),
    )
    assert result.scalars().first() is not None

    # Recreate tables
    await create_db_and_tables(force=True)

    # Final phase: Verify data is gone
    result = await session.execute(select(DBThread).where(DBThread.id == "test-thread"))
    assert result.scalars().first() is None

    result = await session.execute(
        select(DBMessage).where(DBMessage.thread_id == "test-thread"),
    )
    assert result.scalars().first() is None


async def test_relationship_operations(session):
    """Test database operations with relationships."""
    # Create thread and messages
    thread = DBThread(id="test-thread")
    session.add(thread)
    await session.commit()
    await session.refresh(thread)

    message1 = DBMessage(
        thread_id=thread.id,
        message={"role": "user", "content": "test1", "kind": "request"},
    )
    message2 = DBMessage(
        thread_id=thread.id,
        message={"role": "assistant", "content": "test2", "kind": "response"},
    )
    session.add(message1)
    session.add(message2)
    await session.commit()

    # Test relationship loading
    result = await session.execute(
        select(DBThread)
        .where(DBThread.id == "test-thread")
        .options(selectinload(DBThread.messages))
    )
    loaded_thread = result.scalar_one()
    assert len(loaded_thread.messages) == 2
    assert {
        loaded_thread.messages[0].message["content"],
        loaded_thread.messages[1].message["content"],
    } == {"test1", "test2"}

    # Delete messages first
    for message in loaded_thread.messages:
        await session.delete(message)
    await session.commit()
    await session.refresh(loaded_thread)

    # Then delete thread
    await session.delete(loaded_thread)
    await session.commit()

    # Verify all gone
    result = await session.execute(
        select(DBMessage).where(DBMessage.thread_id == "test-thread"),
    )
    assert result.scalars().first() is None

    result = await session.execute(select(DBThread).where(DBThread.id == "test-thread"))
    assert result.scalars().first() is None



if __name__ == "__main__":
    import uvicorn
    from settings import settings

    uvicorn.run(
        "api:app",
        host=settings.host,
        port=settings.port,
        reload=settings.test_mode,
        reload_dirs=["examples/slackbot"] if settings.test_mode else None,
    )



import pytest

import marvin


def test_agent_memory():
    a = marvin.Agent(memories=[marvin.Memory(key="numbers")])
    marvin.run("remember the number 123", agents=[a])
    result = marvin.run("what number did you remember?", agents=[a])
    assert "123" in result


def test_task_memory():
    t = marvin.Task("remember the number 123", memories=[marvin.Memory(key="numbers")])
    t2 = marvin.Task(
        "what number did you remember?",
        memories=[marvin.Memory(key="numbers")],
    )

    t.run()
    result = t2.run()
    assert "123" in result


def test_instructions():
    m = marvin.Memory(
        key="colors",
        instructions="when remembering a color, always store it and the word 'house' e.g. 'red house'",
    )
    a = marvin.Agent(memories=[m])
    marvin.run("remember the color green", agents=[a])
    result = marvin.run("what exactly did you remember?", agents=[a])
    assert "green" in result and "house" in result


@pytest.mark.flaky(reruns=2)
def test_use_memory_as_tool():
    m = marvin.Memory(key="colors")
    a = marvin.Agent(memories=[m])
    marvin.run("remember the color green", agents=[a])
    with marvin.Thread() as t:
        result = marvin.run("what color did you remember?", agents=[a])

    assert "green" in result

    # --- check tool call ---
    messages = t.get_messages()
    found_tool_call = False
    for message in messages:
        for part in message.message.parts:
            if (
                part.part_kind == "tool-call"
                and part.tool_name == "search_memories__colors"
            ):
                found_tool_call = True
                break
        if found_tool_call:
            break
    assert found_tool_call, "Expected to find a tool call to search_memories__colors"


def test_autouse_memory():
    m = marvin.Memory(key="colors", auto_use=True)
    a = marvin.Agent(memories=[m])
    marvin.run("remember the color green", agents=[a])
    with marvin.Thread() as t:
        result = marvin.run("what color did you remember?", agents=[a])

    assert "green" in result

    # --- check tool call did NOT happen---
    messages = t.get_messages()
    found_tool_call = False
    for message in messages:
        for part in message.message.parts:
            if (
                part.part_kind == "tool-call"
                and part.tool_name == "search_memories__colors"
            ):
                found_tool_call = True
                break
        if found_tool_call:
            break
    assert not found_tool_call, (
        "Expected to not find a tool call to search_memories__colors since it is auto-used"
    )



from typing import Any

import marvin
from marvin.agents.agent import Agent
from marvin.handlers.handlers import AsyncHandler, Handler
from marvin.thread import Thread
from marvin.utilities.asyncio import run_sync

DEFAULT_PROMPT = """
You are an expert summarizer that distills information into clear, concise summaries
while preserving the most important semantic meaning and key points. Examine the 
provided `data`, text, or information and create a summary that captures the essence
of the content.

Guidelines for summarization:
- Maintain objectivity and accuracy
- Preserve key facts, figures, and relationships
- Focus on the most important information
- Use clear, concise language
- Adapt tone and style to match the content type
- Ensure the summary stands alone as a coherent piece of text"""

PROMPT = DEFAULT_PROMPT  # for backwards compatibility


async def summarize_async(
    data: Any,
    instructions: str | None = None,
    agent: Agent | None = None,
    thread: Thread | str | None = None,
    context: dict[str, Any] | None = None,
    handlers: list[Handler | AsyncHandler] | None = None,
    prompt: str | None = None,
) -> str:
    """Asynchronously creates a summary of the input data using a language model.

    This function uses a language model to analyze the input data and create a
    concise summary that captures the key information and semantic meaning.

    Args:
        data: The input data to summarize. Can be any type.
        instructions: Optional additional instructions to guide the summarization.
            Used to provide specific guidance about what aspects to focus on or
            how to structure the summary.
        agent: Optional custom agent to use for summarization. If not provided,
            the default agent will be used.
        thread: Optional thread for maintaining conversation context. Can be
            either a Thread object or a string thread ID.
        context: Optional dictionary of additional context to include in the task.
        handlers: Optional list of handlers to use for the task.
        prompt: Optional prompt to use for the task. If not provided, the default
            prompt will be used.

    Returns:
        A string containing the generated summary.

    Examples:
        >>> # Basic summarization
        >>> await summarize_async("Long article about climate change...")
        'Key findings on climate impacts...'

        >>> # Summarize with specific instructions
        >>> await summarize_async("Technical documentation...", instructions="Focus on API changes")
        'Major API updates include...'

    """
    task_context = context or {}
    task_context["Data to summarize"] = data
    prompt = prompt or DEFAULT_PROMPT
    if instructions:
        prompt += (
            f"\n\nYou must follow these instructions for your summary:\n{instructions}"
        )

    task = marvin.Task[str](
        name="Summarize Task",
        instructions=prompt,
        context=task_context,
        result_type=str,
        agents=[agent] if agent else None,
    )

    return await task.run_async(thread=thread, handlers=handlers)


def summarize(
    data: Any,
    instructions: str | None = None,
    agent: Agent | None = None,
    thread: Thread | str | None = None,
    context: dict[str, Any] | None = None,
    handlers: list[Handler | AsyncHandler] | None = None,
    prompt: str | None = None,
) -> str:
    """Creates a summary of the input data using a language model.

    This function uses a language model to analyze the input data and create a
    concise summary that captures the key information and semantic meaning.

    Args:
        data: The input data to summarize. Can be any type.
        instructions: Optional additional instructions to guide the summarization.
            Used to provide specific guidance about what aspects to focus on or
            how to structure the summary.
        agent: Optional custom agent to use for summarization. If not provided,
            the default agent will be used.
        thread: Optional thread for maintaining conversation context. Can be
            either a Thread object or a string thread ID.
        context: Optional dictionary of additional context to include in the task.
        handlers: Optional list of handlers to use for the task.
        prompt: Optional prompt to use for the task. If not provided, the default
            prompt will be used.

    Returns:
        A string containing the generated summary.

    Examples:
        >>> # Basic summarization
        >>> summarize("Long article about climate change...")
        'Key findings on climate impacts...'

        >>> # Summarize with specific instructions
        >>> summarize("Technical documentation...", instructions="Focus on API changes")
        'Major API updates include...'

    """
    return run_sync(
        summarize_async(
            data=data,
            instructions=instructions,
            agent=agent,
            thread=thread,
            context=context,
            handlers=handlers,
            prompt=prompt,
        ),
    )



"""Tasks for Marvin.

A Task is a container for a prompt and its associated state.
"""

import enum
import json
import uuid
from collections.abc import Callable
from contextvars import ContextVar
from dataclasses import dataclass, field
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Annotated,
    Any,
    AsyncGenerator,
    Generic,
    Literal,
    Optional,
    Sequence,
    TypeAlias,
    TypeVar,
)

from pydantic import TypeAdapter
from pydantic_ai.messages import UserContent

import marvin
import marvin.thread
from marvin.agents.actor import Actor
from marvin.agents.team import Swarm, Team
from marvin.memory.memory import Memory
from marvin.prompts import Template
from marvin.thread import Thread
from marvin.utilities.asyncio import run_sync
from marvin.utilities.types import Labels, as_classifier, is_classifier

if TYPE_CHECKING:
    from marvin.engine.end_turn import EndTurn
    from marvin.engine.events import Event
    from marvin.handlers.handlers import AsyncHandler, Handler

T = TypeVar("T")
NOTSET: Literal["__NOTSET__"] = "__NOTSET__"

# Global context var for current task
_current_task: ContextVar[Optional["Task[Any]"]] = ContextVar(
    "current_task",
    default=None,
)

_type_adapters: dict[type[Any], TypeAdapter[Any]] = {}


def get_type_adapter(result_type: type[T]) -> TypeAdapter[T]:
    if result_type not in _type_adapters:
        _type_adapters[result_type] = TypeAdapter(result_type)
    return _type_adapters[result_type]


class TaskState(str, enum.Enum):
    """State of a task."""

    PENDING = "pending"
    RUNNING = "running"
    SUCCESSFUL = "successful"
    FAILED = "failed"
    SKIPPED = "skipped"


ResultType: TypeAlias = type[T] | Annotated[Any, Any] | Labels | Literal["__NOTSET__"]


@dataclass(kw_only=True, init=False)
class Task(Generic[T]):
    """A task is a container for a prompt and its associated state."""

    name: str | None = field(
        default=None,
        metadata={"description": "Optional name for this task"},
    )

    instructions: str = field(
        metadata={"description": "Instructions for the task"},
        kw_only=False,
    )

    attachments: Sequence[UserContent] = field(
        default_factory=list,
        metadata={"description": "Attachments to the task"},
    )

    result_type: ResultType[T] = field(  # type: ignore[reportRedeclaration]
        default=NOTSET,
        metadata={
            "description": "The expected type of the result. This can be a type or None if no result is expected. If not set, the result type will be str.",
        },
        kw_only=False,
    )

    id: str = field(
        default_factory=lambda: uuid.uuid4().hex[:8],
        metadata={"description": "Unique identifier for this task"},
        init=False,
        repr=False,
    )

    prompt_template: str | Path = field(
        default=Path("task.jinja"),
        metadata={
            "description": "Optional Jinja template for customizing how the task appears in prompts. Will be rendered with a `task` variable containing this task instance.",
        },
        repr=False,
    )

    actor: Actor | None = field(
        default=None,
        metadata={
            "description": "The actor assigned to this task. If not provided, the default agent will be used.",
        },
    )

    context: dict[str, Any] = field(
        default_factory=dict,
        metadata={"description": "Context for the task"},
    )

    tools: list[Callable[..., Any]] = field(
        default_factory=list,
        metadata={
            "description": "Tools to make available to any agents assigned to this task",
        },
    )

    memories: list[Memory] = field(
        default_factory=list,
        metadata={
            "description": "Memories to make available to any agents assigned to this task",
        },
    )

    state: TaskState = field(
        default=TaskState.PENDING,
        metadata={"description": "Current state of the task"},
        init=False,
    )

    result_validator: Callable[..., Any] | None = field(
        default=None,
        metadata={
            "description": "Optional function that validates the result. Takes the raw result and returns a validated result or raises an error.",
        },
        repr=False,
    )

    verbose: bool = field(
        default=False,
        metadata={
            "description": "Verbose tasks print additional information to the thread, such as the fact that they started or completed.",
        },
    )

    _parent: "Task[Any] | None" = field(
        default=None,
        metadata={"description": "Optional parent task"},
        init=False,
        repr=False,
    )

    subtasks: set["Task[Any]"] = field(
        default_factory=set,
        metadata={
            "description": "List of subtasks, or tasks for which this task is the parent"
        },
        init=False,
        repr=False,
    )

    depends_on: set["Task[Any]"] = field(
        default_factory=set,
        metadata={
            "description": "List of tasks that must be completed before this task can be run"
        },
        repr=False,
    )

    result: T | str | None = field(
        default=None,
        metadata={
            "description": "The result of the task. Can be either the expected type T or an error string.",
        },
        init=False,
        repr=False,
    )

    allow_fail: bool = field(
        default=False,
        metadata={"description": "Whether to allow the task to fail"},
        repr=False,
    )
    allow_skip: bool = field(
        default=False,
        metadata={"description": "Whether to allow the task to skip"},
        repr=False,
    )

    cli: bool = field(
        default=False,
        metadata={
            "description": "If True, agents will be given a tool for interacting with users on the CLI.",
        },
    )

    # Add _tokens field for context management
    _tokens: list[Any] = field(default_factory=list, init=False, repr=False)

    def __init__(
        self,
        instructions: str | Sequence[UserContent],
        attachments: Sequence[UserContent] | None = None,
        result_type: ResultType[T] = NOTSET,
        *,
        name: str | None = None,
        prompt_template: str | Path = Path("task.jinja"),
        agents: list[Actor | Team] | None = None,
        context: dict[str, Any] | None = None,
        tools: list[Callable[..., Any]] | None = None,
        memories: list[Memory] | None = None,
        result_validator: Callable[..., Any] | None = None,
        parent: "Task[Any] | None | Literal['__NOTSET__']" = NOTSET,
        depends_on: Sequence["Task[Any]"] | None = None,
        allow_fail: bool = False,
        allow_skip: bool = False,
        verbose: bool = False,
        cli: bool = False,
        plan: bool = False,
    ) -> None:
        """Initialize a Task.

        Args:
            instructions: Instructions for the task
            attachments: Optional attachments to the task
            name: Optional name for this task
            result_type: Expected type of the result
            prompt_template: Optional Jinja template for customizing task appearance
            agents: Optional list of agents to execute this task. If more than
                one agent or team is provided, they will automatically be combined
                into a team.
            context: Context for the task
            tools: Tools to make available to agents
            memories: Memories to make available to agents
            result_validator: Optional function to validate results
            parent: Optional parent task. If not provided and there is a current task
                in the context, that task will be used as the parent.
            allow_fail: Whether to allow the task to fail
            allow_skip: Whether to allow the task to skip
            verbose: Whether to print additional status messages to the active thread
            cli: Whether to enable CLI interaction tools
            plan: Whether to enable a tool for planning subtasks

        """
        # required fields
        if isinstance(instructions, str):
            self.instructions = instructions
        else:
            str_instructions = []
            attachments = attachments or []
            for i in instructions:
                if isinstance(i, str):
                    str_instructions.append(i)
                else:
                    attachments.append(i)
            self.instructions = "\n\n".join(str_instructions)

        # optional fields with defaults
        self.attachments = attachments or []
        self.name = name
        self.result_type = result_type if result_type is not NOTSET else str
        self.prompt_template = prompt_template
        self.context = context or {}
        self.tools = tools or []
        self.memories = memories or []
        self.result_validator = result_validator
        self.allow_fail = allow_fail
        self.allow_skip = allow_skip
        self.cli = cli
        self.plan = plan
        self.verbose = verbose
        # key fields
        self.id = uuid.uuid4().hex[:8]
        self.state = TaskState.PENDING
        self.result = None

        # if no parent is provided, use the current task from context
        if parent is NOTSET:
            parent = _current_task.get()
        self.parent = parent
        self.subtasks: set[Task[Any]] = set()
        self.depends_on: set[Task[Any]] = set(depends_on or [])

        # internal fields
        self._tokens = []

        if agents:
            if len(agents) == 1:
                self.actor = agents[0]
            else:
                self.actor = Swarm(members=agents)

        # Handle result type validation
        if isinstance(self.result_type, (list, tuple, set)):
            if len(self.result_type) == 1 and isinstance(
                self.result_type[0],
                (list, tuple, set),
            ):
                if not self.result_type[0]:
                    raise ValueError(
                        "Empty nested list is not allowed for multi-label classification",
                    )
                self.result_type = Labels(self.result_type[0], many=True)
            else:
                if not self.result_type:
                    raise ValueError("Empty list is not allowed for classification")
                if any(isinstance(x, (list, tuple, set)) for x in self.result_type):
                    raise ValueError(
                        "Invalid nested list format - use [['a', 'b']] for multi-label",
                    )
                self.result_type = Labels(self.result_type)

    def __hash__(self) -> int:
        return hash(self.id)

    def _validate_result_type(self) -> None:
        """Validates the result type by converting classification shorthand into
        Labels and ensuring that the result type is a valid type.

        Valid shorthand:
        - result_type = ["red", "blue"] -> Labels(values=["red", "blue"])
        - result_type = [["red", "blue"]] -> Labels(values=["red", "blue"], many=True)
        """
        # Handle double-nested list shorthand for multi-label
        if (
            isinstance(self.result_type, list)
            and len(self.result_type) == 1
            and isinstance(self.result_type[0], (list, tuple, set))
        ):
            if not self.result_type[0]:
                raise ValueError(
                    "Empty nested list is not allowed for multi-label classification",
                )
            self.result_type = Labels(self.result_type[0], many=True)
        # Handle raw sequences for single-label
        elif isinstance(self.result_type, (list, tuple, set)):
            if not self.result_type:
                raise ValueError("Empty list is not allowed for classification")
            if any(isinstance(x, (list, tuple, set)) for x in self.result_type):
                raise ValueError(
                    "Invalid nested list format - use [['a', 'b']] for multi-label",
                )
            self.result_type = Labels(self.result_type)

    def friendly_name(self, verbose: bool = True) -> str:
        """Get a friendly name for this task."""
        if self.name:
            return f'Task {self.id} ("{self.name}")'
        if verbose:
            # Replace consecutive newlines with a single space
            instructions = " ".join(self.instructions.split())
            return f'Task {self.id} ("{instructions[:40]}...")'
        else:
            return f"Task {self.id}"

    @property
    def parent(self) -> "Task[Any] | None":
        """Get the parent task of this task."""
        return self._parent

    @parent.setter
    def parent(self, value: "Task[Any] | None") -> None:
        """Set the parent task of this task."""
        if self._parent is not None:
            self._parent.subtasks.discard(self)
        self._parent = value
        if self._parent is not None:
            self._parent.subtasks.add(self)

    def get_actor(self) -> Actor:
        """Retrieve the actor assigned to this task."""
        return self.actor or marvin.defaults.agent

    def get_tools(self) -> list[Callable[..., Any]]:
        """Get the tools assigned to this task."""
        tools: list[Callable[..., Any]] = []
        tools.extend(self.tools)
        tools.extend([t for m in self.memories for t in m.get_tools()])
        if self.cli:
            import marvin.tools.interactive.cli

            tools.append(marvin.tools.interactive.cli.cli)

        return tools

    def is_classifier(self) -> bool:
        """Return True if this task is a classification task."""
        return is_classifier(self.result_type)

    def get_result_type(self) -> type[T]:
        """Get the effective result type for this task.
        For classification tasks, returns the type that should be used
        for validation (e.g., int or list[int]).
        """
        if self.is_classifier():
            return as_classifier(self.result_type).get_type()
        return self.result_type

    def get_result_type_str(self) -> str:
        """Get a string representation of the result type."""
        if self.is_classifier():
            return (
                f"Provide the integer indices of your chosen labels: "
                f"{as_classifier(self.result_type).get_indexed_labels()}"
            )
        else:
            type_adapter = get_type_adapter(self.get_result_type())
            try:
                return json.dumps(type_adapter.json_schema())

            except Exception:
                return str(self.get_result_type())

    def validate_result(self, raw_result: Any) -> T:
        """Validate a result against the expected type and custom validator."""
        # Apply custom validation if provided
        if self.result_validator is not None:
            try:
                return self.result_validator(raw_result)
            except Exception as e:
                raise ValueError(f"Error validating task result: {e}")

        # Handle classification types
        if self.is_classifier():
            return as_classifier(self.result_type).validate(raw_result)

        result_type = self.get_result_type()

        if result_type is None:
            if raw_result is None:
                return None
            else:
                raise ValueError("Result type is None but result is not None")
        elif raw_result is None:
            raise ValueError("Result is None but result type is not None")

        type_adapter = get_type_adapter(result_type)
        return type_adapter.validate_python(raw_result)

    def get_prompt(self) -> str | Sequence[UserContent]:
        """Get the rendered prompt for this task.

        Uses the task's prompt_template (or default if None) and renders it with
        this task instance as the `task` variable.
        """
        return [
            Template(source=self.prompt_template).render(task=self),
            *self.attachments,
        ]

    async def run_async(
        self,
        *,
        thread: Thread | str | None = None,
        raise_on_failure: bool = True,
        handlers: list["Handler | AsyncHandler"] | None = None,
    ) -> T:
        await marvin.fns.run.run_tasks_async(
            [self],
            thread=thread,
            raise_on_failure=raise_on_failure,
            handlers=handlers,
        )
        return self.result

    def run(
        self,
        *,
        thread: Thread | str | None = None,
        raise_on_failure: bool = True,
        handlers: list["Handler | AsyncHandler"] | None = None,
    ) -> T:
        return run_sync(
            self.run_async(
                thread=thread, raise_on_failure=raise_on_failure, handlers=handlers
            ),
        )

    def run_stream(
        self,
        *,
        thread: Thread | str | None = None,
        raise_on_failure: bool = True,
        handlers: list["Handler | AsyncHandler"] | None = None,
    ) -> AsyncGenerator["Event", None]:
        return marvin.fns.run.run_tasks_stream(
            [self],
            thread=thread,
            raise_on_failure=raise_on_failure,
            handlers=handlers,
        )

    def get_end_turn_tools(self) -> list[type["EndTurn"]]:
        """Get the result tool for this task."""
        import marvin.engine.end_turn

        tools: list[type[marvin.engine.end_turn.EndTurn]] = []
        tools.append(self.mark_successful_tool())
        if self.allow_fail:
            tools.append(self.mark_failed_tool())
        if self.allow_skip:
            tools.append(self.mark_skipped_tool())
        if self.plan:
            tools.append(marvin.engine.end_turn.create_plan_subtasks(parent_task=self))

        return tools

    def mark_successful_tool(self) -> type["marvin.engine.end_turn.MarkTaskSuccessful"]:
        import marvin.engine.end_turn

        return marvin.engine.end_turn.create_mark_task_successful(self)

    def mark_failed_tool(self) -> type["marvin.engine.end_turn.MarkTaskFailed"]:
        import marvin.engine.end_turn

        return marvin.engine.end_turn.create_mark_task_failed(self)

    def mark_skipped_tool(self) -> type["marvin.engine.end_turn.MarkTaskSkipped"]:
        import marvin.engine.end_turn

        return marvin.engine.end_turn.create_mark_task_skipped(self)

    # ------ State Management ------

    async def mark_successful(
        self,
        result: T = None,
        validate_result: bool = True,
        thread: Thread | None = None,
    ) -> None:
        """Mark the task as successful with an optional result."""
        if validate_result:
            result = self.validate_result(result)
        self.result = result
        self.state = TaskState.SUCCESSFUL
        if thread is None:
            thread = marvin.thread.get_current_thread()

        if thread and self.verbose:
            await thread.add_info_message_async(
                f"{self.friendly_name()} successful with result {result}",
                prefix="Task state updated",
            )

    async def mark_failed(self, error: str, thread: Thread | None = None) -> None:
        """Mark the task as failed with an error message."""
        self.result = error
        self.state = TaskState.FAILED
        if thread is None:
            thread = marvin.thread.get_current_thread()

        if thread and self.verbose:
            await thread.add_info_message_async(
                f"{self.friendly_name()} failed with error {error}",
                prefix="Task state updated",
            )

    async def mark_running(
        self,
        thread: Thread | None = None,
    ) -> None:
        """Mark the task as running."""
        self.state = TaskState.RUNNING

        if thread is None:
            thread = marvin.thread.get_current_thread()

        if thread:
            await thread.add_user_message_async(
                self.get_prompt(),
            )

    async def mark_skipped(self, thread: Thread | None = None) -> None:
        """Mark the task as skipped."""
        self.state = TaskState.SKIPPED
        if thread is None:
            thread = marvin.thread.get_current_thread()

        if thread and self.verbose:
            await thread.add_info_message_async(
                f"{self.friendly_name()} skipped",
                prefix="Task state updated",
            )

    def is_pending(self) -> bool:
        """Check if the task is pending."""
        return self.state == TaskState.PENDING

    def is_running(self) -> bool:
        """Check if the task is running."""
        return self.state == TaskState.RUNNING

    def is_successful(self) -> bool:
        """Check if the task is successful."""
        return self.state == TaskState.SUCCESSFUL

    def is_failed(self) -> bool:
        """Check if the task is failed."""
        return self.state == TaskState.FAILED

    def is_skipped(self) -> bool:
        """Check if the task is skipped."""
        return self.state == TaskState.SKIPPED

    def is_incomplete(self) -> bool:
        """Check if the task is incomplete."""
        return self.state in (TaskState.PENDING, TaskState.RUNNING)

    def is_complete(self) -> bool:
        """Check if the task is complete."""
        return self.state in (TaskState.SUCCESSFUL, TaskState.FAILED, TaskState.SKIPPED)

    def is_ready(self) -> bool:
        """Check if the task is ready to run.

        A task is ready if it is incomplete and all of its dependencies (including subtasks) are complete.
        """
        return self.is_incomplete() and all(
            t.is_complete() for t in (self.depends_on | self.subtasks)
        )

    def __enter__(self):
        """Set this task as the current task in context."""
        token = _current_task.set(self)
        self._tokens.append(token)
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any):
        """Reset the current task in context."""
        if self._tokens:  # Only reset if we have tokens
            _current_task.reset(self._tokens.pop())



import logging
import platform
import sys
from functools import partial
from pathlib import Path

import pydantic_ai
import typer
from pydantic_core import to_json
from rich.console import Console
from rich.table import Table
from typer import Context, Exit

import marvin

from .dev import dev_app
from .migrations import migrations as migrations_app
from .migrations import migrations_dev

console = Console()

app = typer.Typer(no_args_is_help=True)

# Add the migrations commands to the dev app
dev_app.add_typer(migrations_dev, name="db")

# Add the main app commands
app.add_typer(dev_app, name="dev")
app.add_typer(migrations_app, name="db")
logging.basicConfig(stream=sys.stderr, level=logging.WARNING, force=True)


@app.command()
def version(ctx: Context):
    if ctx.resilient_parsing:
        return

    info = {
        "Marvin version": marvin.__version__,
        "Pydantic AI version": pydantic_ai.__version__,
        "Python version": platform.python_version(),
        "Platform": platform.platform(),
        "Path": f"~/{Path(__file__).resolve().parents[3].relative_to(Path.home())}",
    }

    g = Table.grid(padding=(0, 1))
    g.add_column(style="bold", justify="left")
    g.add_column(style="cyan", justify="right")
    for k, v in info.items():
        g.add_row(k + ":", str(v).replace("\n", " "))
    console.print(g)

    raise Exit()


@app.command()
def x(
    operation: str = typer.Option(
        "extract", "--operation", "-o", help="Operation to perform"
    ),
    target_type: str = typer.Option(
        "str",
        "--type",
        "-t",
        help="Type of entities to extract (any type that can be eval'd)",
    ),
    instructions: str = typer.Option(
        None, "--instructions", "-i", help="Instructions for extraction"
    ),
    n: int = typer.Option(1, "--n", "-n", help="Number of results to generate"),
):
    """
    Extract entities from stdin input.

    Example: echo "one, two, three" | marvin x -t int | jq
    """
    from marvin.settings import settings

    operation_fn = getattr(marvin, operation)
    if n != 1 and operation != "generate":
        raise ValueError("can only specify n for 'generate' operation")

    # Read from stdin if available
    if not sys.stdin.isatty():
        data = sys.stdin.read().strip()
    else:
        print("Error: No input provided via stdin", file=sys.stderr)
        sys.exit(1)

    try:
        target_type = eval(target_type)
    except Exception:
        print(f"Error: Unsupported type '{target_type}'", file=sys.stderr)
        sys.exit(1)

    if settings.log_level == "DEBUG":
        raise Exception(
            "This can't be run with DEBUG logging, preface this command with MARVIN_LOG_LEVEL=CRITICAL"
        )

    try:
        if operation == "extract":
            operation_fn = partial(
                operation_fn, data=data, target=target_type, instructions=instructions
            )
        elif operation == "cast":
            operation_fn = partial(operation_fn, data=data, target=target_type)
        elif operation == "generate":
            operation_fn = partial(
                operation_fn,
                target=target_type,
                n=n,
                instructions=f"use this as context: {data}",
            )
        print(to_json(operation_fn()).decode("utf-8"))
    except AttributeError:
        print(f"Error: Unsupported operation '{operation}'", file=sys.stderr)
        sys.exit(1)
    except Exception as e:
        print(f"Error during extraction: {e}", file=sys.stderr)
        sys.exit(1)


# this callback only exists to force `version` to be called as `marvin
# version` instead of as the default command, which is the default behavior when
# there's only one command. It can be deleted if/when more commands are added.
@app.callback()
def callback():
    pass


if __name__ == "__main__":
    app()



import logging
from functools import lru_cache
from typing import Any

from rich.logging import RichHandler

import marvin


def maybe_quote(value: Any) -> str:
    if isinstance(value, str):
        return f'"{value}"'
    return str(value)


@lru_cache
def get_logger(name: str | None = None) -> logging.Logger:
    """Retrieves a logger with the given name, or the root logger if no name is given.

    Args:
        name: The name of the logger to retrieve.

    Returns:
        The logger with the given name, or the root logger if no name is given.

    Example:
        Basic Usage of `get_logger`
        ```python
        from marvin.utilities.logging import get_logger

        logger = get_logger("marvin.test")
        logger.info("This is a test") # Output: marvin.test: This is a test

        debug_logger = get_logger("marvin.debug")
        debug_logger.debug_kv("TITLE", "log message", "green")
        ```

    """
    parent_logger = logging.getLogger("marvin")

    if name:
        # Append the name if given but allow explicit full names e.g. "marvin.test"
        # should not become "marvin.marvin.test"
        if not name.startswith(parent_logger.name + "."):
            logger = parent_logger.getChild(name)
        else:
            logger = logging.getLogger(name)
    else:
        logger = parent_logger

    return logger


def setup_logging(level: str | None = None) -> None:
    logger = get_logger()

    if level is not None:
        logger.setLevel(level)
    else:
        logger.setLevel(marvin.settings.log_level)

    logger.handlers.clear()

    handler = RichHandler(rich_tracebacks=True, markup=False)
    formatter = logging.Formatter("%(name)s: %(message)s")
    handler.setFormatter(formatter)

    logger.addHandler(handler)
    logger.propagate = False



from enum import Enum

import pytest
from pydantic import BaseModel

import marvin

sentiment = ["Negative", "Positive"]


class GitHubIssueTag(Enum):
    BUG = "bug"
    FEATURE = "feature"
    ENHANCEMENT = "enhancement"
    DOCS = "docs"


class TestList:
    def test_classify_sentiment(self):
        result = marvin.classify("This is a great feature!", sentiment)
        assert result == "Positive"

    def test_classify_negative_sentiment(self):
        result = marvin.classify(
            "This feature is absolutely terrible!",
            sentiment,
        )
        assert result == "Negative"

    def classify_bug_tag(self):
        result = marvin.classify(
            "This is a bug",
            ["bug", "feature", "enhancement", "docs"],
        )
        assert result == "bug"

    def test_classify_number(self):
        # a version of the prompt would choose the label *number* that
        # matched the data, rather than the label *description*
        result = marvin.classify(0, ["letter", "number"])
        assert result == "number"

    def test_classify_object(self):
        """Test that objects are returned from classify"""

        class Person(BaseModel):
            name: str
            age: int

        p1 = Person(name="Alice", age=30)
        p2 = Person(name="Bob", age=25)
        p3 = Person(name="Charlie", age=35)

        result = marvin.classify("a person in wonderland", [p1, p2, p3])
        assert result is p1


class TestEnum:
    def test_classify_bug_tag(self):
        result = marvin.classify("This is a bug", GitHubIssueTag)
        assert result == GitHubIssueTag.BUG

    def test_classify_feature_tag(self):
        result = marvin.classify("This is a great feature!", GitHubIssueTag)
        assert result == GitHubIssueTag.FEATURE

    def test_classify_enhancement_tag(self):
        result = marvin.classify("This is an enhancement", GitHubIssueTag)
        assert result == GitHubIssueTag.ENHANCEMENT

    def test_classify_docs_tag(self):
        result = marvin.classify("This is a documentation update", GitHubIssueTag)
        assert result == GitHubIssueTag.DOCS


class TestBool:
    def test_classify_true(self):
        result = marvin.classify("2+2=4", bool)
        assert result is True

    def test_classify_false(self):
        result = marvin.classify("2+2=5", bool)
        assert result is False

    def test_classify_with_instructions(self):
        result = marvin.classify(
            "This feature is terrible!",
            bool,
            instructions="Is the sentiment positive?",
        )
        assert result is False

    def test_classify_trueish(self):
        result = marvin.classify(
            "y",
            bool,
            instructions="map the input to true/false",
        )
        assert result is True

    def test_classify_falseish(self):
        result = marvin.classify(
            "nope",
            bool,
            instructions="map the input to true/false",
        )
        assert result is False


class TestInstructions:
    def test_classify_positive_sentiment_with_instructions(self):
        result = marvin.classify(
            "This is a great feature!",
            sentiment,
            instructions="It's opposite day.",
        )
        assert result == "Negative"


class TestAsync:
    async def test_classify_positive_sentiment(self):
        result = await marvin.classify_async("This is a great feature!", bool)
        assert result is True


class TestExamples:
    @pytest.mark.flaky(max_runs=3)
    def test_hogwarts_sorting_hat(self):
        description = "Brave, daring, chivalrous -- it's Harry Potter!"

        house = marvin.classify(
            description,
            labels=["Gryffindor", "Hufflepuff", "Ravenclaw", "Slytherin"],
            instructions="You're the sorting hat. Select the house for the student",
        )

        assert house == "Gryffindor"

    @pytest.mark.parametrize(
        "user_input, expected_selection",
        [
            ("I want to do an event with marvin!", "events and relations"),
            ("Well FooCo offered me a better deal", "sales"),
            ("*angry noises*", "support"),
        ],
    )
    def test_call_routing(self, user_input: str, expected_selection: str):
        class Department(Enum):
            SALES = "sales"
            SUPPORT = "support"
            EVENTS = "events and relations"

        def router(transcript: str) -> Department:
            return marvin.classify(
                transcript,
                labels=Department,
                instructions="Select the best department for the customer request",
            )

        assert router(user_input).value == expected_selection


class TestConvertInputData:
    def test_convert_input_data(self):
        class Name(BaseModel):
            first: str
            last: str

        result = marvin.classify(
            Name(first="Alice", last="Smith"),
            ["Alice", "Bob"],
        )
        assert result == "Alice"


class TestMultiLabel:
    def test_multi_label_list(self):
        """Test multi-label classification with a list."""
        result = marvin.classify(
            "This is a red and blue car",
            ["red", "green", "blue"],
            multi_label=True,
        )
        assert result == ["red", "blue"]

    def test_multi_label_enum(self):
        """Test multi-label classification with an enum."""
        result = marvin.classify(
            "This report contains a bug and a documentation update",
            GitHubIssueTag,
            multi_label=True,
        )
        assert result == [GitHubIssueTag.BUG, GitHubIssueTag.DOCS]

    def test_multi_label_bool(self, gpt_4o):
        """Test multi-label classification with a boolean."""
        result = marvin.classify(
            ["2+2=4", "2+2=5"],
            bool,
            multi_label=True,
        )
        assert result == [True, False]



import asyncio
from typing import Literal

from pydantic import BaseModel, Field

import marvin


class AllergyFormatA(BaseModel):
    substance: str = Field(description="Name of the allergic substance")
    severity: str = Field(description="Severity level (mild, moderate, severe)")
    reaction: str = Field(description="Type of allergic reaction")
    dateIdentified: str = Field(description="Date allergy was identified (YYYY-MM-DD)")


class AllergyFormatB(BaseModel):
    agent: str = Field(description="Name of the allergic substance")
    riskLevel: Literal["low", "medium", "high"] = Field(description="Risk level")
    manifestations: list[str] = Field(
        description="List of observed reactions",
        alias="clinicalManifestations",
    )
    documentation_date: str = Field(
        description="Date allergy was documented (YYYY-MM-DD)",
        alias="documentationDate",
    )


async def main():
    allergy_record_a = AllergyFormatA(
        substance="penicillin",
        severity="severe",
        reaction="anaphylaxis",
        dateIdentified="2020-01-01",
    )

    allergy_record_b = await marvin.cast_async(
        allergy_record_a,
        AllergyFormatB,
        instructions="Convert Epic allergy record to Cerner format. Map 'severe' severity to 'high' risk. Format reaction as a list.",
    )
    print(f"Converted to format B: {allergy_record_b}")


if __name__ == "__main__":
    asyncio.run(main())
"""
» python examples/hello_cast.py
╭─ Marvin ─────────────────────────────────────────────────────────────────────────────────────────╮
│ ⠋   Final Result                                                                                 │
│     Input:   {                                                                                   │
│                'response': {                                                                     │
│                  'task_id': '895be677-29f6-4642-ac99-d63171a46444',                              │
│                  'result': {                                                                     │
│                    'agent': 'penicillin',                                                        │
│                    'riskLevel': 'high',                                                          │
│                    'clinicalManifestations': [                                                   │
│                      'anaphylaxis'                                                               │
│                    ],                                                                            │
│                    'documentationDate': '2020-01-01'                                             │
│                  }                                                                               │
│                }                                                                                 │
│              }                                                                                   │
╰──────────────────────────────────────────────────────────────────────────────────── 12:23:47 AM ─╯
Converted to format B: agent='penicillin' riskLevel='high' manifestations=['anaphylaxis'] documentation_date='2020-01-01'
"""



"""Command line utilities for database migrations."""

import asyncio
import os
import sys

import typer
from rich.console import Console
from rich.table import Table

from marvin.database import (
    ALEMBIC_DIR,
    ALEMBIC_INI,
    get_async_engine,
    is_sqlite,
)
from marvin.utilities.logging import get_logger

logger = get_logger(__name__)
console = Console()

# Create a Typer app for user-facing migration commands
migrations = typer.Typer(
    name="db", help="Manage database migrations.", no_args_is_help=True
)

# Create a Typer app for developer migration commands
migrations_dev = typer.Typer(
    help="Developer database migration commands.", no_args_is_help=True
)


def get_alembic_cfg():
    """Get the Alembic config."""
    from alembic.config import Config

    cfg = Config(str(ALEMBIC_INI))
    return cfg


async def _run_revision_with_connection(connection, alembic_cfg, message, autogenerate):
    """Run revision command with async connection."""
    from alembic import command

    def do_revision(connection):
        alembic_cfg.attributes["connection"] = connection
        command.revision(alembic_cfg, message=message, autogenerate=autogenerate)

    await connection.run_sync(do_revision)


@migrations_dev.command("revision")
def revision(
    autogenerate: bool = typer.Option(
        False,
        "--autogenerate/--no-autogenerate",
        help="Automatically generate migrations based on schema changes",
    ),
    message: str = typer.Option(None, "--message", "-m", help="Migration message"),
):
    """Create a new migration revision. Developer command."""
    try:
        # Ensure versions directory exists
        versions_dir = ALEMBIC_DIR / "versions"
        os.makedirs(versions_dir, exist_ok=True)

        alembic_cfg = get_alembic_cfg()

        from alembic import command

        command.revision(alembic_cfg, message=message, autogenerate=autogenerate)
        console.print("[green]Migration revision created[/green]")
        return

    except Exception as e:
        console.print(f"[red]Failed to create migration revision: {e}[/red]")
        sys.exit(1)


@migrations.command("upgrade")
def upgrade(
    revision: str = typer.Argument(
        "head", help="Revision to upgrade to. Use 'head' for latest."
    ),
):
    """Upgrade database to the latest migration."""
    try:
        from alembic import command

        # Ensure versions directory exists
        versions_dir = ALEMBIC_DIR / "versions"
        os.makedirs(versions_dir, exist_ok=True)

        alembic_cfg = get_alembic_cfg()
        command.upgrade(alembic_cfg, revision)
        console.print(f"[green]Database upgraded to {revision}[/green]")
    except Exception as e:
        console.print(f"[red]Failed to upgrade database: {e}[/red]")
        sys.exit(1)


@migrations.command("downgrade")
def downgrade(
    revision: str = typer.Argument(..., help="Revision to downgrade to"),
    yes: bool = typer.Option(
        False, "--yes", "-y", help="Confirm the destructive operation without prompting"
    ),
):
    """Downgrade database to specified revision. Requires confirmation."""
    if not yes:
        confirmation = typer.confirm(
            f"WARNING: You are about to downgrade the database to revision '{revision}'. "
            f"This operation is destructive and may result in data loss. Continue?",
            default=False,
        )
        if not confirmation:
            console.print("[yellow]Operation cancelled.[/yellow]")
            sys.exit(0)

    try:
        from alembic import command

        alembic_cfg = get_alembic_cfg()
        command.downgrade(alembic_cfg, revision)
        console.print(f"[green]Database downgraded to {revision}[/green]")
    except Exception as e:
        console.print(f"[red]Failed to downgrade database: {e}[/red]")
        sys.exit(1)


@migrations.command("reset")
def reset(
    yes: bool = typer.Option(
        False, "--yes", "-y", help="Confirm the destructive operation without prompting"
    ),
):
    """Reset database by downgrading to base and upgrading to head. Requires confirmation."""
    if not yes:
        confirmation = typer.confirm(
            "WARNING: You are about to reset the database to the latest migration. "
            "This operation is destructive and may result in data loss. Continue?",
            default=False,
        )
        if not confirmation:
            console.print("[yellow]Operation cancelled.[/yellow]")
            sys.exit(0)

    try:
        from alembic import command

        alembic_cfg = get_alembic_cfg()

        # Downgrade to base (before first migration)
        console.print("[blue]Downgrading database to base state...[/blue]")
        command.downgrade(alembic_cfg, "base")

        # Upgrade to latest
        console.print("[blue]Upgrading database to latest migration...[/blue]")
        command.upgrade(alembic_cfg, "head")

        console.print("[green]Database reset complete[/green]")
    except Exception as e:
        console.print(f"[red]Failed to reset database: {e}[/red]")
        sys.exit(1)


@migrations.command("history")
def history():
    """Show migration history."""
    try:
        from alembic import command

        alembic_cfg = get_alembic_cfg()
        command.history(alembic_cfg)
    except Exception as e:
        console.print(f"[red]Failed to show migration history: {e}[/red]")
        sys.exit(1)


@migrations.command("current")
def current():
    """Show current migration revision."""
    try:
        from alembic import command

        alembic_cfg = get_alembic_cfg()
        command.current(alembic_cfg)
    except Exception as e:
        console.print(f"[red]Failed to show current migration: {e}[/red]")
        sys.exit(1)


async def _get_migration_status(connection, alembic_cfg, script):
    """Get migration status using async connection."""
    from alembic.runtime.migration import MigrationContext

    def get_current_rev(connection):
        context = MigrationContext.configure(connection)
        return context.get_current_revision()

    return await connection.run_sync(get_current_rev)


@migrations.command("status")
def status():
    """Show database migration status and information."""
    from marvin.settings import settings

    is_sqlite_db = is_sqlite()
    db_url = settings.database_url or "Not configured"

    # Format sensitive parts of the URL for display
    if db_url != "Not configured":
        from urllib.parse import urlparse, urlunparse

        parsed = urlparse(db_url)

        # For better security, create a sanitized URL that masks credentials
        # but preserves the database type, host, port, and database name
        if parsed.netloc and "@" in parsed.netloc:
            # Extract the host:port part after the @ symbol
            host_part = parsed.netloc.split("@")[1]
            # Create a netloc with masked credentials
            masked_netloc = f"****:****@{host_part}"

            # Reconstruct URL with masked credentials
            sanitized_parts = list(parsed)
            sanitized_parts[1] = masked_netloc  # Replace netloc
            display_url = urlunparse(sanitized_parts)
        else:
            # For SQLite or URLs without credentials, use as is
            display_url = db_url
    else:
        display_url = db_url

    # Use a borderless table for config
    config_table = Table.grid(padding=(0, 2))
    config_table.add_column(style="bold")
    config_table.add_column()

    console.print("[bold]DATABASE CONFIGURATION[/bold]")
    config_table.add_row("Database URL:", display_url)
    config_table.add_row(
        "Database Type:", "SQLite" if is_sqlite_db else "PostgreSQL/Other"
    )
    console.print(config_table)
    console.print()

    # Check migration status
    try:
        from alembic.script import ScriptDirectory

        # Ensure versions directory exists
        versions_dir = ALEMBIC_DIR / "versions"
        os.makedirs(versions_dir, exist_ok=True)

        alembic_cfg = get_alembic_cfg()
        script = ScriptDirectory.from_config(alembic_cfg)

        # Get head revision
        head_revision = script.get_current_head()

        # Get current revision using async engine
        async def get_async_status():
            engine = get_async_engine()
            async with engine.connect() as connection:
                current_revision = await _get_migration_status(
                    connection, alembic_cfg, script
                )
            await engine.dispose()
            return current_revision

        current_revision = asyncio.run(get_async_status())

        # Use a borderless table for migration status
        migration_table = Table.grid(padding=(0, 2))
        migration_table.add_column(style="bold")
        migration_table.add_column()

        console.print("[bold]MIGRATION STATUS[/bold]")

        # Show current revision
        current_rev_display = (
            current_revision
            if current_revision
            else "[red]Not set (no migrations applied)[/red]"
        )
        migration_table.add_row("Current Revision:", current_rev_display)

        # Show head revision
        head_rev_display = head_revision if head_revision else "No migrations found"
        migration_table.add_row("Head Revision:", head_rev_display)

        # Determine migration status
        status_text = ""
        status_style = "green"

        if not head_revision:
            status_text = "No migrations exist yet"
            status_style = "yellow"
        elif not current_revision:
            status_text = "Database has no migrations applied but migrations exist"
            status_style = "red"
        elif current_revision == head_revision:
            status_text = "Database is up to date"
            status_style = "green"
        else:
            # Calculate how many revisions behind
            revs = list(script.iterate_revisions(head_revision, current_revision))
            migration_count = len(revs)
            status_text = f"Database is behind by {migration_count} migration{'s' if migration_count > 1 else ''}"
            status_style = "yellow"

        migration_table.add_row(
            "Status:", f"[{status_style}]{status_text}[/{status_style}]"
        )
        console.print(migration_table)

        # Show upgrade message if needed
        if status_style == "yellow" and current_revision != head_revision:
            console.print()
            console.print(
                "[yellow]Run 'marvin db upgrade' to apply pending migrations[/yellow]"
            )

    except Exception as e:
        console.print(f"[yellow]Could not determine migration status: {e}[/yellow]")
        if "No such revision" in str(e):
            console.print(
                "Run 'marvin dev db revision -m \"initial\"' to create an initial migration"
            )



"""Base models for Marvin."""

from pydantic import BaseModel, ConfigDict


class MarvinModel(BaseModel):
    """Base model for all Marvin models.

    This model:
    - Is strict (no extra fields allowed)
    - Uses modern Pydantic v2 features
    """

    model_config = ConfigDict(
        extra="forbid",  # no extra fields allowed
        frozen=False,  # allow mutation
        validate_assignment=True,  # validate on attribute assignment
        validate_default=True,  # validate default values
    )



from dataclasses import dataclass

import marvin


@dataclass
class Currency:
    name: str
    symbol: str


if __name__ == "__main__":
    currencies = marvin.extract(
        "After flying ORD to Heathrow, I exchanged my greenbacks for local currency.",
        target=Currency,
    )
    print(currencies)
"""
» python examples/hello_extract.py
╭─ Marvin ─────────────────────────────────────────────────────────────────────────────────────────╮
│ ⠋   Final Result                                                                                 │
│     Input:   {                                                                                   │
│                'response': {                                                                     │
│                  'task_id': '32669832-4ebc-482f-bcda-dbd2a5f2ea3d',                              │
│                  'result': [                                                                     │
│                    {                                                                             │
│                      'name': 'US Dollar',                                                        │
│                      'symbol': 'USD'                                                             │
│                    },                                                                            │
│                    {                                                                             │
│                      'name': 'British Pound Sterling',                                           │
│                      'symbol': 'GBP'                                                             │
│                    }                                                                             │
│                  ]                                                                               │
│                }                                                                                 │
│              }                                                                                   │
╰──────────────────────────────────────────────────────────────────────────────────── 11:45:43 PM ─╯
[Currency(name='US Dollar', symbol='USD'), Currency(name='British Pound Sterling', symbol='GBP')]
"""



import enum
import inspect
import textwrap
from collections.abc import Callable, Sequence
from dataclasses import dataclass, field
from typing import (
    TYPE_CHECKING,
    Annotated,
    Any,
    Generic,
    Literal,
    ParamSpec,
    TypeAlias,
    TypeVar,
    get_args,
    get_origin,
)

from marvin.utilities.asyncio import run_sync
from marvin.utilities.jinja import jinja_env

T = TypeVar("T")
P = ParamSpec("P")
R = TypeVar("R")

# A type that can be used as a target for type conversion
TargetType: TypeAlias = type[T] | Annotated[T, Any]


@dataclass
class Labels:
    """A container for classification labels.

    This class provides a consistent interface for working with labels,
    whether they come from enums, literals, lists, or other sources.

    Args:
        values: The label values. Can be an enum class, a sequence of values,
            or a Literal type.
        many: Whether this is a multi-label classifier (i.e., can select
            multiple values).

    Examples:
        >>> # Single-label classification with raw values
        >>> labels = Labels(["red", "green", "blue"])
        >>> labels.values
        ("red", "green", "blue")

        >>> # Multi-label classification with raw values
        >>> labels = Labels(["red", "green", "blue"], many=True)
        >>> labels.values
        ("red", "green", "blue")

        >>> # Single-label classification with enum
        >>> class Colors(enum.Enum):
        ...     RED = "red"
        ...     GREEN = "green"
        ...     BLUE = "blue"
        >>> labels = Labels(Colors)
        >>> labels.values  # Returns enum members
        (<Colors.RED: 'red'>, <Colors.GREEN: 'green'>, <Colors.BLUE: 'blue'>)

        >>> # Multi-label classification with enum
        >>> labels = Labels(Colors, many=True)
        >>> labels.values  # Returns enum members
        (<Colors.RED: 'red'>, <Colors.GREEN: 'green'>, <Colors.BLUE: 'blue'>)

    """

    values: type[enum.Enum] | Sequence[Any] | Any
    many: bool = False

    def __post_init__(self):
        # Convert values to a tuple of labels
        if self.values is bool:
            # For bool, always use True first then False
            self._labels = (True, False)
        elif issubclass_safe(self.values, enum.Enum):
            self._labels = tuple(self.values)  # Returns enum members
        elif get_origin(self.values) is Literal:
            self._labels = get_args(self.values)
        elif isinstance(self.values, (list, tuple, set)):
            self._labels = tuple(self.values)  # type: ignore
        else:
            raise ValueError(f"Invalid label type: {type(self.values)}")

    @property
    def labels(self) -> tuple[Any, ...]:
        """Get the label values."""
        return self._labels

    def get_type(self) -> type:
        """Get the type that should be used for validation."""
        return list[int] if self.many else int

    def validate(self, value: int | list[int] | None) -> Any | list[Any]:
        """Validate a value against the labels."""
        if value is None:
            raise ValueError("None is not a valid value for classification")

        if self.many:
            if not isinstance(value, (list, tuple)):
                raise ValueError(
                    f"Expected a list of indices for multi-label classification, got {type(value)}",
                )
            if not value:
                raise ValueError(
                    "Empty list is not allowed for multi-label classification",
                )
            if not all(isinstance(i, int) for i in value):  # type: ignore[reportUnnecessaryIsinstance]
                raise ValueError("All elements must be integers")
            if not all(0 <= i < len(self._labels) for i in value):
                raise ValueError(
                    f"All indices must be between 0 and {len(self._labels) - 1}",
                )
            if len(set(value)) != len(value):
                raise ValueError("Duplicate indices are not allowed")

            # Map indices to actual values
            return [self._labels[i] for i in value]

        # Single label case
        if not isinstance(value, int):
            raise ValueError(
                f"Expected an integer index for classification, got {type(value)}",
            )
        if not (0 <= value < len(self._labels)):
            raise ValueError(
                f"Invalid index {value}. Must be between 0 and {len(self._labels) - 1}",
            )
        return self._labels[value]

    def get_indexed_labels(self) -> dict[int, str]:
        """Get a mapping of indices to label string representations."""

        def format_value(v: Any) -> str:
            if isinstance(v, enum.Enum):
                return repr(v.value)  # Show the enum's value, not its name
            if isinstance(v, str):
                return f"'{v}'"  # Single quotes for strings
            return str(v)

        return {i: format_value(v) for i, v in enumerate(self._labels)}


def as_classifier(type_: type[T]) -> Labels:
    """Convert a type to a Labels instance.
    This should only be called on types that have been verified as classifiers via is_classifier().

    Args:
        typ: A type that represents a classifier (Enum, Literal, sequence, or list thereof)

    Returns:
        Labels: A Labels instance representing the classifier

    Raises:
        ValueError: If the type is not a valid classifier

    """
    typ = type_
    if isinstance(typ, Labels):
        return typ

    # Handle list[T] case for type-level classifiers
    origin = get_origin(typ)
    if origin is list:
        arg = get_args(typ)[0]
        # Handle list[Enum] or list[Literal]
        if (isinstance(arg, type) and issubclass(arg, enum.Enum)) or get_origin(
            arg,
        ) is Literal:
            if TYPE_CHECKING:  # TODO: revisit this
                assert isinstance(arg, type)
            return Labels(arg, many=True)

    # Handle double-nested list shorthand
    if (
        isinstance(typ, list)
        and len(typ) == 1
        and isinstance(typ[0], (list, tuple, set))
    ):
        return Labels(typ[0], many=True)  # type: ignore

    # Convert raw sequences to Labels
    if isinstance(typ, (list, tuple, set)):
        return Labels(typ)

    # Handle remaining single-label cases (Enum, Literal)
    return Labels(typ)


def is_classifier(type_: type[T]) -> bool:
    """Check if a type represents a classification task.
    This includes:
    - Single-label: Enum, Literal, or any sequence of values
    - Multi-label: list[Enum], list[Literal], or list[list]

    Examples:
        >>> class Colors(enum.Enum):
        ...     RED = "red"
        ...     GREEN = "green"
        >>> is_classifier(Colors)  # enum
        True
        >>> is_classifier(Literal["a", "b"])  # literal
        True
        >>> is_classifier(["a", "b"])  # list of values
        True
        >>> is_classifier([1, "red", MyClass()])  # mixed values
        True
        >>> is_classifier(list[Colors])  # multi-label enum
        True
        >>> is_classifier(list[Literal["a", "b"]])  # multi-label literal
        True
        >>> is_classifier(list[["a", 1, MyClass()]])  # multi-label shorthand
        True

    """
    typ = type_
    if isinstance(typ, Labels):
        return True

    # Handle list[T] case
    origin = get_origin(typ)
    if origin is list:
        arg = get_args(typ)[0]
        # Check for list[Enum], list[Literal], or list[list]
        return (
            issubclass_safe(arg, enum.Enum)
            or get_origin(arg) is Literal
            or isinstance(arg, (list, tuple, set))
        )

    # Handle single-label cases
    return (
        # Enum type
        issubclass_safe(typ, enum.Enum)
        # Literal type
        or get_origin(typ) is Literal
        # Any sequence of values
        or isinstance(typ, (list, tuple, set))
    )


def issubclass_safe(x: Any, cls: type | tuple[type, ...]) -> bool:
    """Safely check if x is a subclass of cls without raising errors.

    This combines isinstance(x, type) and issubclass(x, cls) checks in a safe way
    that won't raise TypeError if x is not a type.

    Args:
        x: The value to check
        cls: A type or tuple of types to check against

    Returns:
        bool: True if x is a type and is a subclass of cls, False otherwise

    Example:
        >>> issubclass_safe(str, object)  # type is subclass
        True
        >>> issubclass_safe(42, object)  # not a type
        False
        >>> issubclass_safe(str, (int, float))  # not a subclass
        False

    """
    if isinstance(x, type) and issubclass(x, cls):
        return True
    elif origin := get_origin(x):
        return issubclass_safe(origin, cls)
    return False


@dataclass
class ParameterModel:
    name: str
    annotation: str | None
    default: str | None


@dataclass
class PythonFunction(Generic[P, R]):
    """A dataclass representing a Python function.

    Attributes:
        function (Callable): The original function object.
        signature (inspect.Signature): The signature object of the function.
        name (str): The name of the function.
        docstring (Optional[str]): The docstring of the function.
        parameters (List[ParameterModel]): The parameters of the function.
        return_annotation (Optional[Any]): The return annotation of the function.
        source_code (str): The source code of the function.
        bound_parameters (dict[str, Any]): The parameters of the function bound with values.
        return_value (Optional[Any]): The return value of the function call.

    """

    function: Callable[P, R]
    signature: inspect.Signature
    name: str
    parameters: list[ParameterModel]
    docstring: str | None = None
    return_annotation: Any | None = None
    source_code: str | None = None
    bound_parameters: dict[str, Any] = field(default_factory=dict)
    return_value: Any | None = None

    @property
    def definition(self) -> str:
        docstring = inspect.cleandoc(self.docstring) if self.docstring else ""
        formatted_docstring = textwrap.indent(
            f'"""\n{docstring}\n"""' if docstring else "",
            prefix="    ",
        )
        return f"def {self.name}{self.signature}:\n{formatted_docstring}".strip()

    @classmethod
    def from_function(
        cls,
        func: Callable[P, R],
        **kwargs: Any,
    ) -> "PythonFunction[P, R]":
        """Create a PythonFunction instance from a function.

        Args:
            func (Callable): The function to create a PythonFunction instance from.
            **kwargs: Additional keyword arguments to set as attributes on the PythonFunction instance.

        Returns:
            PythonFunction: The created PythonFunction instance.

        """
        name = kwargs.pop("name", func.__name__)
        docstring = kwargs.pop("docstring", func.__doc__)
        sig = inspect.signature(func)
        parameters = [
            ParameterModel(
                name=name,
                annotation=(
                    str(param.annotation)
                    if param.annotation is not param.empty
                    else None
                ),
                default=(
                    repr(param.default) if param.default is not param.empty else None
                ),
            )
            for name, param in sig.parameters.items()
        ]

        try:
            source_code = inspect.getsource(func).strip()
        except OSError as e:
            error_message = str(e)
            if "source code" in error_message:
                source_code = None
            else:
                raise

        function_dict: dict[str, Any] = {
            "function": func,
            "signature": sig,
            "name": name,
            "docstring": inspect.cleandoc(docstring) if docstring else None,
            "parameters": parameters,
            "return_annotation": sig.return_annotation,
            "source_code": source_code,
        }

        function_dict.update(kwargs)

        return cls(**function_dict)

    @classmethod
    def from_function_call(
        cls,
        func: Callable[P, R],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> "PythonFunction[P, R]":
        """Create a PythonFunction instance from a function call.

        Args:
            func (Callable): The function to call.
            *args: Positional arguments to pass to the function call.
            **kwargs: Keyword arguments to pass to the function call.

        Returns:
            PythonFunction: The created PythonFunction instance, with the return value of the function call set as an attribute.

        """
        sig = inspect.signature(func)

        bound = sig.bind(*args, **kwargs)
        bound.apply_defaults()

        return_value = func(*bound.args, **bound.kwargs)
        if inspect.iscoroutine(return_value):
            return_value = run_sync(return_value)

        # render the docstring with the bound arguments, if it was supplied as jinja
        docstring = jinja_env.from_string(func.__doc__ or "").render(
            **dict(bound.arguments.items()),
        )

        instance = cls.from_function(
            func,
            docstring=docstring,
            bound_parameters={k: v for k, v in bound.arguments.items()},
            return_value=return_value,
        )

        return instance
